# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import os
from textwrap import dedent


def bool_env(varname: str, default: bool) -> bool:
    """Read an environment variable and interpret it as a boolean.
    True: 'y', 'yes', 't', 'true', 'on', and '1';
    False: 'n', 'no', 'f', 'false', 'off', and '0'.

    Args:
        varname: the name of the variable
        default: the default boolean value
    """
    val = os.getenv(varname, str(default))
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif val in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        raise ValueError(f"invalid truth value {val!r} for environment {varname!r}")


def int_env(varname: str, default: int) -> int:
    """Read an environment variable and interpret it as an integer."""
    return int(os.getenv(varname, default))


def get_env(varname: str, type, default: int | bool) -> int | bool:
    if type is int:
        return int_env(varname, int(default))
    elif type is bool:
        return bool_env(varname, bool(default))
    else:
        raise TypeError(f"Unknown type {type}")


class ReadOnlyDict:
    def __init__(self, data):
        self._data = data

    def __getitem__(self, key):
        return self._data[key]


class Config:
    _HAS_DYNAMIC_ATTRIBUTES = True

    def __init__(self):
        object.__setattr__(self, "_values", {})
        object.__setattr__(self, "_types", {})
        object.__setattr__(self, "_editable_at_runtime", {})
        object.__setattr__(self, "_meta", {})

        object.__setattr__(self, "_readonly", ReadOnlyDict(self._values))
        object.__setattr__(self, "_callbacks", {})

    def define(
        self,
        name,
        type,
        default,
        *,
        help,
        runtime=False,
        callback=None,
        lazy=False,
    ):  # noqa: W0613
        """
        Defines a new flag

        Args:
            name: the flag name, should be an uppercase string like "NETKET_XXX"
            type: should be the type (bool, int) of the flag
            default: default value
            help: a string to use as description of this flag
            runtime: whether the flag can be modified at runtime
            callback: an optional callback function taking the value as argument
                that is executed when the flag is changed
            lazy: do not call the callback at init.
        """
        if name in self._values:
            raise KeyError(f"Flag {name} already defined.")

        self._types[name] = type
        self._editable_at_runtime[name] = runtime
        self._values[name] = get_env(name, type, default)
        self._callbacks[name] = callback
        self._meta[name] = (
            type,
            [],
            {
                "help": help,
            },
        )

        if callback is not None and not lazy:
            callback(self._values[name])

    @property
    def FLAGS(self):
        """
        The flags of this instance of netket
        """
        return self._readonly

    def update(self, name, value):
        """
        Updates a configuration variable in netket.

        Args:
            name: the name of the variable
            value: the new value
        """
        name = name.upper()

        if not self._editable_at_runtime[name]:
            raise RuntimeError(
                f"\n\nFlag `{name}` can only be set through an environment "
                "variable before importing netket.\n"
                "Try launching python with:\n\n"
                f"\t{name}={self.FLAGS[name]} python\n\n"
                "or execute the following snippet BEFORE importing netket:\n\n"
                "\t>>>import os\n"
                f'\t>>>os.environ["{name}"]="{self.FLAGS[name]}"\n'
                "\t>>>import netket as nk\n\n"
            )

        if not isinstance(value, self._types[name]):
            raise TypeError(
                f"Configuration {name} must be a {self._types[name]}, but the "
                f"value {value} is a {type(value)}."
            )

        self._values[name] = self._types[name](value)

    def __repr__(self):
        txt = "\nGlobal configurations for NetKet\n"
        for k, v in self._values.items():
            txt = txt + f" - {k} = {v}\n"

        return txt

    def __dir__(self):
        return list(k.lower() for k in self._values.keys())

    def __getattr__(self, name: str) -> Any:
        """Handle dynamically created attributes."""
        if name == name.lower():
            upper_name = name.upper()
            if upper_name in self._values:
                return self.FLAGS[upper_name]
        raise AttributeError(f"'Config' object has no attribute '{name}'")

    def __setattr__(self, name: str, value: Any) -> None:
        """Handle setting dynamically created attributes."""
        if name == name.lower() and name.upper() in self._values:
            # Handle config flag setting
            upper_name = name.upper()
            if self._callbacks[upper_name] is not None:
                self._callbacks[upper_name](value)
            self.update(upper_name, value)
            return
        # Use default behavior for everything else
        super().__setattr__(name, value)


config = Config()
FLAGS = config.FLAGS

config.define(
    "NETKET_DEBUG",
    bool,
    default=False,
    help="Enable debug logging in many netket functions.",
    runtime=True,
)

config.define(
    "NETKET_EXPERIMENTAL",
    bool,
    default=False,
    help="Enable experimental features.",
    runtime=True,
)

config.define(
    "NETKET_USE_PLAIN_RHAT",
    bool,
    default=False,
    help=dedent(
        """
        By default, NetKet uses the split-R̂ Gelman-Rubin diagnostic in `netket.stats.statistics`,
        which detects non-stationarity in the MCMC chains (in addition to the classes of
        chain-mixing failures detected by plain R) since version 3.4.
        Enabling this flag restores the previous behavior of using plain (non-split) Rhat.
        """
    ),
    runtime=True,
)

config.define(
    "NETKET_EXPERIMENTAL_FFT_AUTOCORRELATION",
    bool,
    default=False,
    help=dedent(
        """
        The integrated autocorrelation time $\tau_c$ is computed separately for each chain $c$.
        To summarize it for the user, `Stats.tau_corr` is changed to contain the average over all
        chains and a new field `Stats.tau_corr_max` is added containing the maximum autocorrelation
        among all chains (which helps to identify outliers). Using the average $\tau$ over all chains
        seems like a good choice as it results in a low-variance estimate
        (see [here](https://emcee.readthedocs.io/en/stable/tutorials/autocorr/#autocorr) for a good
        discussion).
        """
    ),
    runtime=True,
)


def _setup_experimental_sharding_cpu(n_procs):
    if n_procs > 1:
        import jax

        jax.config.update("jax_num_cpu_devices", n_procs)


config.define(
    "NETKET_EXPERIMENTAL_SHARDING_CPU",
    int,
    default=0,
    help=dedent(
        """
        Set to >=1 to force JAX to use multiple threads as separate devices on cpu.
        Sets the XLA_FLAGS='--xla_force_host_platform_device_count=#' environment variable.
        Disabled by default.
        """
    ),
    runtime=False,
    callback=_setup_experimental_sharding_cpu,
)


def _update_x64(val):
    from jax import config as jax_config

    jax_config.update("jax_enable_x64", val)


# This flag is setup to mirror JAX_ENABLE_X64 with True default. any of the two
# Can be explicitly set.
config.define(
    "NETKET_ENABLE_X64",
    bool,
    default=bool_env(
        "JAX_ENABLE_X64", True
    ),  # respect explicit JAX_ENABLE_X64 settings
    help=dedent(
        """
        Enables double-precision for Jax. Equivalent to `JAX_ENABLE_X64` but defaults to
        True instead of False, as it is required throughout NetKet. By setting this flag
        to False NetKet will run without double-precision everywhere.
        """
    ),
    runtime=True,
    callback=_update_x64,
)


config.define(
    "NETKET_SPHINX_BUILD",
    bool,
    default=False,
    help=dedent(
        """
        Set to True when building documentation with Sphinx. Disables some decorators.
        """
    ),
    runtime=True,
)


def _setup_experimental_sharding(val, explicit=False):
    if val:
        import jax
        from jax.sharding import AxisType

        if explicit:
            axis_types = AxisType.Explicit
        else:
            axis_types = AxisType.Auto

        mesh = jax.make_mesh(
            (len(jax.devices()),),
            ("S"),
            axis_types=axis_types,
        )
        jax.sharding.set_mesh(mesh)

        # mode_type = "Explicit" if explicit else "Auto"
        # warnings.warn(
        #     f"""
        #     NETKET_EXPERIMENTAL_SHARDING mode detected:
        #         - SHARDING IS NOW ALWAYS ENABLED, BUT TO USE MORE THAN 1 DEVICE YOU MUST
        #         DEFINE THE MESH AND SPECIFY IT IN YOUR CODE.

        #         - For backward compatibility, specifying `NETKET_EXPERIMENTAL_SHARDING` will
        #         create create and set a single-axis mesh with all devices for you.

        #         - You should UPDATE YOUR CODE to include the following lines, and stop declaring
        #         `NETKET_EXPERIMENTAL_SHARDING` in your code.

        #         import jax
        #         import netket as nk

        #         # Create a mesh with all the devices
        #         mesh = jax.make_mesh(
        #             (len(jax.devices()),),  # How many devices
        #             ("S"),                  # The name of the axis. 'S' is standard for 'samples'.
        #             axis_types=(
        #                 AxisType.{mode_type},  # Explicit/Auto sharding mode
        #             ),
        #         )
        #         jax.sharding.set_mesh(mesh) # Set this as the default mesh for jax.

        #     """
        # )


config.define(
    "NETKET_EXPERIMENTAL_SHARDING",
    bool,
    default=True,
    help=dedent(
        """
        Enables highly expermiental support of netket for running on multiple jax devices.

        Supports both multiple local devices, as well as global ones in a multi-process environment.
        See https://jax.readthedocs.io/en/latest/multi_process.html#initializing-the-cluster for
        how to initialize the latter.
        Distributes chains and samples equally among all available devices.
        """
    ),
    runtime=True,
    callback=_setup_experimental_sharding,
)


config.define(
    "NETKET_RANDOM_STATE_FALLBACK_WARNING",
    bool,
    default=True,
    runtime=True,
    help=dedent(
        """
        Print a warning every time you use a fallback that could never stop running
        when using random_state of constrained hilbert spaces with custom
        constraints.
        """
    ),
)


config.define(
    "NETKET_EXPERIMENTAL_SHARDING_FAST_SERIALIZATION",
    bool,
    default=False,
    runtime=True,
    help=dedent(
        """
        If True (Defaults False) does not gather data on the master process when
        using flax.serialization methods. This allows to use orbax-checkpoint with
        higher efficiency.
        """
    ),
)


config.define(
    "NETKET_NO_TIPS",
    bool,
    default=False,
    runtime=True,
    help=dedent(
        """
        If True, disables the display of random tips when NetKet is imported
        interactively.
        """
    ),
)



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Utility functions for computing traces of permutation operators

import numpy as np

from itertools import product


def count_n_uplets(values, target, m):
    r"""
    Count the number of n-uplets $s \in \{0, ..., m-1\}^n$ such that
    $$ \sum_{k=0}^{n-1} 2 s_k a_k = t $$
    where $a_k$ corresponds to `values[k]` and $t$ to `target`.

    Args:
        values (<int>): The list of n integers.
        target (int): The target value for the sums.
        m (int): The number of choice per index of values.

    Return:
        int: The number of such n-uplets.
    """

    table = np.full((len(values) + 1, target + 1), -1)
    table[0] = 0
    table[0][0] = 1

    for i in range(len(values)):
        for j in range(table.shape[1]):
            accumulated_value = 0
            # Add the contribution from element i being added at magnitude k
            for k in range(m):
                not_included_sum = j - 2 * k * values[i]
                if not_included_sum >= 0:
                    accumulated_value += table[i][not_included_sum]
                table[i + 1][j] = accumulated_value

    return table[-1][-1]


def get_subset_occupations(partition_labels, subsets):
    """
    Given a list of subsets of [0, ..., n-1] and a partition of [0, ..., n-1],
    return a list of the number of element in each partition for each subset.

    Args:
        partition_labels (<int>): The list such that partition_labels[k] is the partition to which k belongs.
        subsets (<<int>>): The list of subsets.

    Return:
        <ndarray>: The list such that l[i][j] is the number of elements of subsets[i] in partition j.
    """
    n_partitions = len(np.unique(partition_labels))
    occupations = []
    for subset in subsets:
        occupation_count = np.zeros(n_partitions, dtype=int)
        for k in subset:
            occupation_count[partition_labels[k]] += 1
        occupations.append(occupation_count)
    return occupations


def get_parity_sum(occupation_list, n_occupations):
    """
    Given a list occupation_list of lists of length p and n_occupations, a list of length p,
    we look at all subsets of occupation_list such that the sum of its elements is n_occupations.
    We return the sum over all such subsets, of the parity of the number of indices k in that subset such
    that sum(occupation_list[k]) is even.

    Args:
        occupation_list (<ndarray>): The list of lists of length p.
        n_occupations (<int>): The list of target occupation.

    Return:
        int: The sum of parities specified above.
    """

    table = np.full(
        (
            len(occupation_list) + 1,
            *(n_occupation + 1 for n_occupation in n_occupations),
        ),
        -1,
    )

    table[0] = 0
    table[0][(0,) * len(n_occupations)] = 1

    for i in range(len(occupation_list)):

        occupation_iterator = product(
            *list(range(n_occupation + 1) for n_occupation in n_occupations)
        )

        for occupation in occupation_iterator:

            not_included_sum = np.array(occupation) - occupation_list[i]
            if np.any(not_included_sum < 0):
                table[i + 1][occupation] = table[i][occupation]
            else:
                table[i + 1][occupation] = (-1) ** (
                    sum(occupation_list[i]) + 1
                ) * table[i][tuple(not_included_sum)] + table[i][occupation]

    return table[-1][tuple(n_occupations)].item()



__all__ = [
    "PermutationOperator",
    "PermutationOperatorFermion",
    "construct_permutation_operator",
]

from netket._src.operator.permutation.spin import (
    PermutationOperator as PermutationOperator,
)
from netket._src.operator.permutation.fermion import (
    PermutationOperatorFermion as PermutationOperatorFermion,
)
from netket._src.operator.permutation.construct import (
    construct_permutation_operator as construct_permutation_operator,
)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Ignore false-positives for redefined `product` functions:
# pylint: disable=function-redefined

import itertools
from typing import overload, Literal, Any

import numpy as np

from netket.utils import HashableArray, struct
from netket.utils.float import comparable, prune_zeros
from netket.utils.types import Array
from netket.utils.dispatch import dispatch

from netket.utils.group._semigroup import Element, FiniteSemiGroup, Identity


@struct.dataclass
class FiniteGroup(FiniteSemiGroup):
    """
    Collection of Elements expected to satisfy group axioms.
    Unlike FiniteSemiGroup, product tables, conjugacy classes, etc. can be calculated.

    Group elements can be implemented in any way, as long as a subclass of Group
    is able to implement their action. Subclasses must implement a :code:`_canonical()`
    method that returns an array of integers for each acceptable Element such that
    two Elements are considered equal iff the corresponding matrices are equal.

    """

    def __hash__(self):
        return super().__hash__()

    def _canonical(self, x: Element) -> Array:
        r"""
        Canonical form of :code:`Element`s, used for equality testing (i.e., two
        :code:`Element`s `x,y` are deemed equal iff
        :code:`_canonical(x) == _canonical(y)`.
        Must be overridden in subclasses

        Arguments:
            x: an `Element`

        Returns:
            the canonical form as a numpy.ndarray of integers
        """
        raise NotImplementedError

    def _canonical_array(self) -> Array:
        r"""
        Lists the canonical forms returned by `_canonical` as rows of a 2D array.
        """
        return np.array([self._canonical(x).flatten() for x in self.elems])

    def _canonical_lookup(self) -> dict:
        r"""
        Creates a lookup table from canonical forms to index in `self.elems`
        """
        return {
            HashableArray(self._canonical(element)): index
            for index, element in enumerate(self.elems)
        }

    @overload
    def remove_duplicates(
        self, *, return_inverse: Literal[False] = False
    ) -> "FiniteGroup": ...

    @overload
    def remove_duplicates(
        self, *, return_inverse: Literal[True]
    ) -> tuple["FiniteGroup", Any]: ...

    def remove_duplicates(self, *, return_inverse=False):
        r"""
        Returns a new :class:`FiniteGroup` with duplicate elements (that is,
        elements with identical canonical forms) removed.

        Arguments:
            return_inverse: If True, also return indices to reconstruct the
                original group from the result.

        Returns:
            The group with duplicate elements removed. If
            :code:`return_inverse==True` it also returns the list of indices
            needed to reconstruct the original group from the result.
        """
        result = np.unique(
            self._canonical_array(),
            axis=0,
            return_index=True,
            return_inverse=return_inverse,
        )
        group = FiniteGroup([self.elems[i] for i in sorted(result[1])])
        if return_inverse:
            return group, result[2]
        else:
            return group

    @struct.property_cached
    def inverse(self) -> Array:
        r"""
        Indices of the inverse of each element.

        Assuming the definitions

        .. code::

            g = self[idx_g]
            h = self[self.inverse[idx_g]]

        :code:`gh = product(g, h)` is equivalent to :code:`Identity()`
        """
        canonical_identity = self._canonical(Identity())
        inverse = np.zeros(len(self.elems), dtype=int)

        for i, e1 in enumerate(self.elems):
            for j, e2 in enumerate(self.elems):
                prod = e1 @ e2
                if np.all(self._canonical(prod) == canonical_identity):
                    inverse[i] = j

        return inverse

    @struct.property_cached
    def product_table(self) -> Array:
        r"""
        A table of indices corresponding to :math:`g^{-1} h` over the group.

        Assuming the definitions

        .. code::

            g = self[idx_g]
            h = self[idx_h]
            idx_u = self.product_table[idx_g, idx_h]

        :code:`self[idx_u]` corresponds to :math:`u = g^{-1} h` .
        """
        n_symm = len(self)
        product_table = np.zeros([n_symm, n_symm], dtype=int)

        lookup = self._canonical_lookup()

        for i, e1 in enumerate(self.elems[self.inverse]):
            for j, e2 in enumerate(self.elems):
                prod = e1 @ e2
                product_table[i, j] = lookup[HashableArray(self._canonical(prod))]

        return product_table

    @struct.property_cached
    def conjugacy_table(self) -> Array:
        r"""
        The conjugacy table of the group.

        Assuming the definitions

        .. code::

            g = self[idx_g]
            h = self[idx_h]

        :code:`self[self.conjugacy_table[idx_g,idx_h]]`
        corresponds to :math:`h^{-1}gh`.
        """
        col_index = np.arange(len(self))[np.newaxis, :]
        # exploits that h^{-1}gh = (g^{-1} h)^{-1} h
        return self.product_table[self.product_table, col_index]

    @struct.property_cached
    def conjugacy_classes(self) -> tuple[Array, Array, Array]:
        r"""
        The conjugacy classes of the group.

        Returns:

            The three arrays

            - classes: a boolean array, each row indicating the elements that
              belong to one conjugacy class
            - representatives: the lowest-indexed member of each conjugacy class
            - inverse: the conjugacy class index of every group element

        """
        row_index = np.arange(len(self))[:, np.newaxis]

        # if is_conjugate[i,j] is True, self[i] and self[j] are in the same class
        is_conjugate = np.full((len(self), len(self)), False)
        is_conjugate[row_index, self.conjugacy_table] = True

        classes, representatives, inverse = np.unique(
            is_conjugate, axis=0, return_index=True, return_inverse=True
        )

        # Usually self[0] == Identity(), which is its own class
        # This class is listed last by the lexicographic order used by np.unique
        # so we reverse it to get a more conventional layout
        classes = classes[::-1]
        representatives = representatives[::-1]
        inverse = (representatives.size - 1) - inverse

        return classes, representatives, inverse

    def check_multiplier(self, multiplier: Array, rtol=1e-8, atol=0) -> bool:
        r"""
        Checks the associativity constraint of Schur multipliers.

        .. math::

            \alpha(x, y) \alpha(xy, z) = \alpha(x, yz) \alpha(y, z).


        Arguments:
            multiplier: the array of Schur multipliers :math:`\alpha(x,y)`
            rtol: relative tolerance
            atol: absolute tolerance

        Returns:
            whether :code:`multiplier` is a valid Schur multiplier
            up to the given tolerance

        Raises:
            :class:`ValueError`
                if the shape of `multiplier` does not match
                the size of the group
        """
        if multiplier.shape != (len(self), len(self)):
            raise ValueError(
                "Schur multipliers must form a square array of the same size\n"
                f"as the group ({len(self)}, {len(self)}), "
                f"got {multiplier.shape} instead!"
            )

        PT = self.product_table[self.inverse]
        # \alpha(x, y) \alpha(xy, z)
        left = np.expand_dims(multiplier, 2) * multiplier[PT, :]
        # \alpha(x, yz) \alpha(y, z)
        right = multiplier[:, PT] * np.expand_dims(multiplier, 0)
        return np.allclose(left, right, rtol=rtol, atol=atol)

    def _character_from_class_matrix(
        self, class_matrix: Array, which_class: Array | None = None
    ) -> np.ndarray:
        r"""Given a linear combination of class matrices,
        diagonalise it and normalise the eigenvectors as characters:

        .. math::

            \sum_g |\chi(g)|^2 = \sum_C |C| |\chi(c_0)|^2 = |G|.

        Arguments:
            class_matrix: square matrix diagonalised by the characters
                :math:`\chi(c_0)` (without any further scaling).
            which_class: (optional) boolean array, specifying which
                conjugacy classes are represented in `class_matrix`.

        Returns:
            irrep characters obtained from the eigenvectors of `class_matrix`.

            Excluded classes are filled in with zeros. The irreps are
            sorted by dimension.
        """
        _, table = np.linalg.eig(class_matrix)
        table = table.T  # want eigenvectors as rows

        # normalisation
        classes, _, _ = self.conjugacy_classes
        class_sizes = classes.sum(axis=1)
        if which_class is not None:
            class_sizes = class_sizes[which_class]
        norm = np.sum(np.abs(table) ** 2 * class_sizes, axis=1, keepdims=True)
        norm = (norm / len(self)) ** 0.5
        table /= norm
        # ensure correct sign (i.e. identity should have a real character)
        table /= _cplx_sign(table[:, 0])[:, np.newaxis]

        # Sort lexicographically, ascending by first column, descending by others
        sorting_table = np.column_stack((table.real, table.imag))
        sorting_table[:, 1:] *= -1
        sorting_table = comparable(sorting_table)
        _, indices = np.unique(sorting_table, axis=0, return_index=True)
        table = table[indices]

        # Get rid of annoying nearly-zero entries
        table = prune_zeros(table)

        if which_class is not None:
            # fill in excluded classes
            full_table = np.zeros((len(table), len(classes)), dtype=table.dtype)
            full_table[:, which_class] = table
            return full_table
        else:
            return table

    def projective_characters_by_class(
        self, multiplier: Array | None
    ) -> tuple[np.ndarray, np.ndarray]:
        r"""
        Calculates the character table of projective representations with a
        given Schur multiplier α using a modified Burnside algorithm.

        Arguments:
            multiplier: the unitary Schur multiplier.
                If unspecified, computes linear representation characters.

        Returns:
            - :code:`characters_by_class`
                a 2D array, each row containing the characters of a
                representative element of each conjugacy class in one
                projective irrep with the given multiplier.
            - :code:`class_factors`
                a 1D array listing the "class factors" of each element of
                the group. The character of each element is the product
                of the character of the class representative with this
                class factor.
                (Only returned if :code:`multiplier` is not :code:`None`.)

        Note: the algorithm and the definitions above are explained in more
        detail in https://arxiv.org/abs/2505.14790.
        """
        # deal with trivial multipliers
        if multiplier is None:
            return self.character_table_by_class
        elif np.allclose(multiplier, 1.0, rtol=0, atol=1e-8):
            # still trivial but should conform to projective irrep return format
            return self.character_table_by_class, np.ones(len(self))

        # check unitarity
        if not np.allclose(np.abs(multiplier), 1.0, rtol=0, atol=1e-8):
            raise ValueError("Schur multiplier must be unitary")

        # compute class factors
        class_factors = np.zeros(len(self), dtype=multiplier.dtype)
        classes, class_reps, _ = self.conjugacy_classes
        reg_classes = np.zeros(len(class_reps), dtype=bool)

        numerator = multiplier[self.inverse, np.arange(len(self))]
        for i, (g, cls) in enumerate(zip(class_reps, classes)):
            # β(g, h^-1 g h) = α(h^-1, h) / α(h^-1, g) α(h^-1 g, h)
            β = (
                numerator
                / multiplier[self.inverse, g]
                / multiplier[self.product_table[:, g], np.arange(len(self))]
            )

            class_size = cls.sum()
            centralizer_size = len(self) // class_size
            # collate sets of h with equal h^-1 g h
            h_sets = np.argsort(self.conjugacy_table[g]).reshape(
                class_size, centralizer_size
            )
            # average β across each set
            β = np.average(β[h_sets], axis=1)

            if np.allclose(np.abs(β), 1.0, rtol=0, atol=1e-8):
                # if the class is α-regular, β for each averaged entry was equal
                # and the array now contains unit complex numbers
                class_factors[cls] = β
                reg_classes[i] = True
            elif not np.allclose(np.abs(β), 0.0, rtol=0, atol=1e-8):
                # otherwise, the different β should average to zero
                raise RuntimeError(
                    "Class factors close to neither unity of zero\n" + repr(β)
                )

        # The algorithm hinges on the equation
        #   \sum_C M_{ABC} \chi(c_0) = |A||B| \chi(a_0) \chi(b_0) / d_\chi,
        # where
        #   M_{ABC} = \sum_{a \in A, b \in B, ab \in C} α(a,b) β(ab)/β(a)β(b)
        #
        # From the oblique times table, we can easily extract the array
        #   (M'_B)_{AC} = \sum_{a \in A, ab_0 \in C} α(a, b_0) β(ab_0)/β(a)
        #               = M_{ABC} / |B|
        # by building an array of α(a, b_0) β(ab_0)/β(a), filtering for
        # `product_table == b_0`, and aggregating over conjugacy classes.
        #
        # Defining
        #   (M_B)_{AC} := (M'_B)_{AC} / |A|,
        # we finally have the eigenvalue equation
        #   M_B \vec\chi = \chi(b_0)/d_\chi \vec\chi,
        # where \vec\chi is the vector of characters \chi(c_0).

        # Construct a random linear combination of the M_B
        # array of α(a,b) β(ab)/β(a)
        class_matrix = (
            multiplier[np.arange(len(self))[:, None], self.product_table]
            * class_factors
            # set non-regular elements to zero instead of inf/nan
            / np.where(class_factors == 0.0, np.inf, class_factors)[:, None]
        )

        # multiply with random weights for α-regular class representatives
        # TODO should we have a term for every α-regular element?
        weight_ = random(reg_classes.sum(), seed=0, cplx=np.iscomplexobj(multiplier))
        weight = np.zeros(len(self), dtype=weight_.dtype)
        weight[class_reps[reg_classes]] = weight_
        class_matrix *= weight[self.product_table]

        # aggregate by α-regular conjugacy class, divide by |A|
        classes = classes[reg_classes]
        class_matrix = classes @ class_matrix @ classes.T
        class_matrix /= classes.sum(axis=1)[:, None]

        return (
            self._character_from_class_matrix(class_matrix, reg_classes),
            class_factors,
        )

    @struct.property_cached
    def character_table_by_class(self) -> np.ndarray:
        r"""
        Calculates the character table using Burnside's algorithm.

        Each row of the output lists the characters of one irrep in the order the
        conjugacy classes are listed in :attr:`conjugacy_classes`.

        Assumes that :code:`Identity() == self[0]`, if not, the sign of some
        characters may be flipped. The irreps are sorted by dimension.
        """
        classes, _, _ = self.conjugacy_classes
        class_sizes = classes.sum(axis=1)
        # Burnside's algorithm hinges on the equation
        #   \sum_C M_{ABC} \chi_C = |A||B| \chi_A \chi_B / d_\chi,
        # where
        #   M_{ABC} = #{a \in A, b \in B: ab \in C}.
        #
        # From the oblique times table, we can easily extract the array
        #   (M'_B)_{AC} = #{a \in A: ab_0 \in C}
        # by aggregating `product_table == b_0` over conjugacy classes.
        #
        # Defining
        #   (M_B)_{AC} := (M'_B)_{AC} / |A|,
        # we finally have the eigenvalue equation
        #   M_B \vec\chi = \chi_B / d_\chi \vec\chi,
        # where \vec\chi is the vector of characters \chi_C.

        # Construct a random linear combination of the M_B
        # i.e. weight each `product_table==b` with some random weight w(b)
        class_matrix = (
            classes @ random(len(self), seed=0)[self.product_table] @ classes.T
        )
        class_matrix /= class_sizes[:, None]

        return self._character_from_class_matrix(class_matrix)

    def character_table(self, multiplier: Array | None = None) -> np.ndarray:
        r"""
        Calculates the character table using Burnside's algorithm.

        Arguments:
            multiplier: (optional) Schur multiplier

        Returns:
            a matrix of all linear irrep characters (if :code:`multiplier is None`)
            or projective irrep characters with the given :code:`multiplier`,
            sorted by dimension.

            Each row of lists the characters of all group elements
            for one irrep, i.e. :code:`self.character_table()[i,g]`
            gives :math:`\chi_i(g)`.

        It is assumed that :code:`Identity() == self[0]`. If not, the sign
        of some characters may be flipped and the sorting by dimension
        will be wrong.
        """
        _, _, class_idx = self.conjugacy_classes
        if multiplier is None or np.allclose(multiplier, 1.0, rtol=0, atol=1e-8):
            # linear representations
            CT = self.character_table_by_class
            return CT[:, class_idx]
        else:
            # projective representations
            CT, class_factor = self.projective_characters_by_class(multiplier)
            return CT[:, class_idx] * class_factor

    def character_table_readable(
        self, multiplier: Array | None = None, full: bool = False
    ) -> tuple[list[str], Array]:
        r"""
        Returns a conventional rendering of the character table.

        Arguments:
            multiplier: (optional) Schur multiplier
            full: whether the character table for all group elements (True)
                or one representative per conjugacy class (False, default)

        Returns:

            A tuple containing a list of strings and an array

            - :code:`classes`: a text description of a representative of
              each conjugacy class (or each group element) as a list
            - :code:`characters`: a matrix, each row of which lists the
              characters of one irrep
        """
        if full:
            names = [str(g) for g in self]
            return names, self.character_table(multiplier)
        else:
            classes, idx_repr, _ = self.conjugacy_classes
            class_sizes = classes.sum(axis=1)
            representatives = [
                f"{class_sizes[cls]}x{self[rep]}" for cls, rep in enumerate(idx_repr)
            ]
            if multiplier is None or np.allclose(multiplier, 1.0, rtol=0, atol=1e-8):
                # linear representations
                CT = self.character_table_by_class
            else:
                # projective representations
                CT, _ = self.projective_characters_by_class(multiplier)
            return representatives, CT

    @struct.property_cached
    def _irrep_matrices(self) -> list[Array]:
        # We use Dixon's algorithm (Math. Comp. 24 (1970), 707) to decompose
        # the regular representation of the group into its irreps.
        # We start with a Hermitian matrix E that commutes with every matrix in
        # this rep: the space spanned by its degenerate eigenvectors then all
        # transform according to some rep of the group.
        # The matrix is randomised to ensure there are no accidental degeneracies,
        # i.e. all these spaces are irreps.
        # For real irreps, real matrices are returned: if needed, the same
        # routine is run with a real symmetric and a complex Hermitian matrix.

        true_product_table = self.product_table[self.inverse]
        inverted_product_table = true_product_table[:, self.inverse]

        def invariant_subspaces(e, seed):
            # Construct a Hermitian matrix that commutes with all matrices
            # in the regular rep.
            # These matrices obey E_{g,h} = e_{gh^{-1}} for some vector e
            e = e[inverted_product_table]
            e = e + e.T.conj()

            # Since E commutes with all the ρ, its eigenspaces reduce the rep
            # For a random input vector, there are no accidental degeneracies
            # except for complex irreps and real symmetric E.
            e, v = np.linalg.eigh(e)

            # indices that split v into eigenspaces
            _, starting_idx = np.unique(comparable(e), return_index=True)

            # Calculate sᴴPv for one eigenvector v per eigenspace, a fixed
            # random vector s and all regular projectors P
            # These are calculated as linear combinations of sᴴρv for the
            # regular rep matrices ρ, which is given by the latter two terms
            vs = v[:, starting_idx]
            s = random(len(self), seed=seed)[inverted_product_table]
            # row #i of this `s` is sᴴρ(self[i]), where sᴴ is the random vector
            proj = self.character_table().conj() @ s @ vs
            starting_idx = list(starting_idx) + [len(self)]
            return v, starting_idx, proj

        # Calculate the eigenspaces for a real and/or complex random matrix
        # To check which are needed, calculate Frobenius-Schur indicators
        # This is +1, 0, -1 for real, complex, and quaternionic irreps
        squares = np.diag(true_product_table)
        frob = np.array(
            np.rint(
                np.sum(self.character_table()[:, squares], axis=1).real / len(self)
            ),
            dtype=int,
        )
        eigen = {}
        if np.any(frob == 1):
            # real irreps: start from a real symmetric invariant matrix
            e = random(len(self), seed=0)
            eigen["real"] = invariant_subspaces(e, seed=1)
        if np.any(frob != 1):
            # complex or quaternionic irreps: complex hermitian invariant matrix
            e = random(len(self), seed=2, cplx=True)
            eigen["cplx"] = invariant_subspaces(e, seed=3)

        irreps = []

        for i, chi in enumerate(self.character_table()):
            v, idx, proj = eigen["real"] if frob[i] == 1 else eigen["cplx"]
            # Check which eigenspaces belong to this irrep
            proj = np.logical_not(np.isclose(proj[i], 0.0))
            # Pick the first eigenspace in this irrep
            first = np.arange(len(idx) - 1, dtype=int)[proj][0]
            v = v[:, idx[first] : idx[first + 1]]
            # v is now the basis of a single irrep: project the regular rep onto it
            irreps.append(np.einsum("gi,ghj ->hij", v.conj(), v[true_product_table, :]))

        return irreps

    def irrep_matrices(self) -> list[Array]:
        """
        Returns matrices that realise all irreps of the group.

        Returns:
            A list of 3D arrays such that :code:`self.irrep_matrices()[i][g]`
            contains the representation of :code:`self[g]` consistent with
            the characters in :code:`self.character_table()[i]`.
        """
        return self._irrep_matrices


def _cplx_sign(x):
    return x / np.abs(x)


def random(n, seed, cplx=False):
    if cplx:
        v = np.random.default_rng(seed).normal(size=(2, n))
        v = v[0] + 1j * v[1]
        return v
    else:
        return np.random.default_rng(seed).normal(size=n)


@dispatch
def product(A: FiniteGroup, B: FiniteGroup):
    return FiniteGroup(
        elems=[a @ b for a, b in itertools.product(A.elems, B.elems)],
    )



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union
from collections.abc import Iterable

from netket.operator._abstract_operator import AbstractOperator
from netket.operator._sum.base import SumOperator


class SumGenericOperator(SumOperator, AbstractOperator):
    def __init__(
        self,
        *operators: AbstractOperator,
        coefficients: Union[float, Iterable[float]] = 1.0,
        dtype=None,
    ):
        if not all(isinstance(op, AbstractOperator) for op in operators):
            raise TypeError(
                "Arguments to SumOperator must all be subtypes of "
                "AbstractOperator. However the types are:\n\n"
                f"{list(type(op) for op in operators)}\n"
            )
        super().__init__(
            operators, operators[0].hilbert, coefficients=coefficients, dtype=dtype
        )



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
from flax import linen as nn
from typing import Any


class ApplyOperatorModuleLinen(nn.Module):
    """
    A Flax Linen module that wraps another module and applies an operator transformation.

    This module wraps a base neural network module and applies an operator in front of it,
    computing log(O|ψ⟩) where O is the operator and |ψ⟩ is represented by the base module.

    The operator is stored in flattened form. The static structure (treedef) is stored as a
    module attribute, while the dynamic data (leaves) is stored in the 'operator' variable
    collection, which is not trainable. This separation allows the operator's arrays to be
    updated without triggering recompilation while keeping the structure static.

    Args:
        base_module: The Flax module to wrap
        operator_treedef: The pytree structure of the operator (obtained from jax.tree.flatten)

    Example::

        import netket as nk
        hilbert = nk.hilbert.Spin(0.5, 4)
        base_module = nk.models.RBM(alpha=1)
        operator = nk.operator.spin.sigmax(hilbert, 0)

        # Flatten the operator to separate static structure from dynamic data
        leaves, treedef = jax.tree.flatten(operator)
        transformed = ApplyOperatorModuleLinen(base_module=base_module, operator_treedef=treedef)

        # Initialize: first init the base module to get its params
        base_params = base_module.init(jax.random.key(1), hilbert.all_states())
        # Then add only the operator leaves to the variables dict
        variables = {**base_params, 'operator': {'leaves': leaves}}

        # Apply the transformed module
        logpsi = transformed.apply(variables, x)

        # The operator can be updated without recompilation
        # Only update the leaves (treedef is fixed in the module)
        new_leaves, _ = jax.tree.flatten(new_operator)
        variables['operator']['leaves'] = new_leaves
        logpsi = transformed.apply(variables, x)
    """

    base_module: nn.Module
    """The base module to wrap"""

    operator_treedef: Any
    """The static pytree structure of the operator"""

    @classmethod
    def from_module_and_variables(
        cls, bare_module: nn.Module, operator, bare_variables: dict
    ):
        """
        Create a TransformedModule from a bare module, operator, and variables.

        Args:
            bare_module: The bare Flax module to wrap
            operator: The operator to apply (will be flattened)
            bare_variables: The variables dictionary from the bare module

        Returns:
            A tuple of (wrapped_module, wrapped_variables) where:
            - wrapped_module is the new TransformedModule instance
            - wrapped_variables is the properly structured variables dict
        """
        leaves, treedef = jax.tree.flatten(operator)
        wrapped_module = cls(base_module=bare_module, operator_treedef=treedef)

        # Wrap the base module's variables under 'base_module' key to match Flax's scoping
        wrapped_variables = {
            collection: {"base_module": params}
            for collection, params in bare_variables.items()
        }

        # Add the new operator at the top level (outer module's operator)
        if "operator" not in wrapped_variables:
            wrapped_variables["operator"] = {}
        wrapped_variables["operator"]["leaves"] = leaves

        return wrapped_module, wrapped_variables

    @property
    def operator(self):
        """Reconstruct the operator from its flattened representation."""
        if not self.has_variable("operator", "leaves"):
            raise ValueError(
                "Operator leaves not found in variables. "
                "Make sure to add the operator leaves to variables:\n"
                "  leaves, treedef = jax.tree.flatten(operator)\n"
                "  variables['operator'] = {'leaves': leaves}"
            )
        leaves = self.get_variable("operator", "leaves")
        return jax.tree.unflatten(self.operator_treedef, leaves)

    @nn.compact
    def __call__(self, x, *args, **kwargs):
        if self.is_initializing():
            raise ValueError(
                "ApplyOperatorModuleLinen cannot be initialized using the standard init() method. "
                "Instead, initialize the base_module separately and provide operator leaves:\n"
                "  base_params = base_module.init(key, x)\n"
                "  leaves, treedef = jax.tree.flatten(operator)\n"
                "  variables = {**base_params, 'operator': {'leaves': leaves}}"
            )

        # TODO: Move to global import to avoid circular dependencies
        from netket.operator import ContinuousOperator

        # Reconstruct the operator using the property
        stored_operator = self.operator

        if isinstance(stored_operator, ContinuousOperator):
            # For continuous operators, use their specialized kernel
            # Create a wrapper function that calls the base module with apply()
            base_variables = {
                k: v for k, v in self.variables.items() if k != "operator"
            }

            def apply_base(variables, x_in, *args_in, **kwargs_in):
                return self.base_module.apply(variables, x_in, *args_in, **kwargs_in)

            res = stored_operator._expect_kernel(
                apply_base, base_variables, x, *args, **kwargs
            )
        else:
            xp, mels = stored_operator.get_conn_padded(x)
            xp = xp.reshape(-1, x.shape[-1])

            # Call the base module on the connected configurations using apply()
            logpsi_xp = self.base_module(xp, *args, **kwargs)
            logpsi_xp = logpsi_xp.reshape(mels.shape).astype(complex)

            res = jax.scipy.special.logsumexp(logpsi_xp, axis=-1, b=mels)

        return res



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Functional approach for transforming apply functions with operators.

This module provides the functional approach for cases where the model
is neither a Linen module nor an NNX module. It wraps an apply function
to apply an operator transformation.
"""

import jax
import flax

from netket import jax as nkjax


def make_logpsi_op_afun(logpsi_fun, operator, variables):
    """Wraps an apply_fun into another one that multiplies it by an operator.

    This wrapper is made such that the operator is passed as the model_state
    of the new wrapped function, and therefore changes to the angles/coefficients
    of the operator should not trigger recompilation.

    Args:
        logpsi_fun: a function that takes as input variables and samples
        operator: a {class}`nk.operator.JaxDiscreteOperator`
        variables: The variables used to call *logpsi_fun*

    Returns:
        A tuple, where the first element is a new function with the same signature as
        the original **logpsi_fun** and a set of new variables to be used to call it.

    """
    # Wrap logpsi into logpsi_op
    logpsi_op_fun = nkjax.HashablePartial(_logpsi_op_fun, logpsi_fun)

    # Insert a new 'operator' key to store the operator. This only works
    # if operator is a pytree that can be flattened/unflattened.
    new_variables = flax.core.copy(variables, {"operator": operator})

    return logpsi_op_fun, new_variables


def _logpsi_op_fun(apply_fun, variables, x, *args, **kwargs):
    """
    This should be used as a wrapper to the original apply function, adding
    to the `variables` dictionary (in model_state) a new key `operator` with
    a jax-compatible operator.
    """
    # TODO: Move to global import
    from netket.operator import ContinuousOperator

    variables_applyfun, operator = flax.core.pop(variables, "operator")

    if isinstance(operator, ContinuousOperator):
        res = operator._expect_kernel(apply_fun, variables_applyfun, x)
    else:
        xp, mels = operator.get_conn_padded(x)
        xp = xp.reshape(-1, x.shape[-1])
        logpsi_xp = apply_fun(variables_applyfun, xp, *args, **kwargs)
        logpsi_xp = logpsi_xp.reshape(mels.shape).astype(complex)

        res = jax.scipy.special.logsumexp(logpsi_xp, axis=-1, b=mels)
    return res



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax

from netket.utils import config


def inspect(name: str, tree: jax.Array):
    """
    Internal function to inspect the sharding of an array. To be used for debugging inside
    of :func:`jax.jit`-ted functions.

    Args:
        name: A string to identify the array, usually the name, but can contain anything else.
        x: The array
    """

    def _inspect(name: str, x: jax.Array):
        if config.netket_experimental_sharding:

            def _cb(y):
                if jax.process_index() == 0:
                    print(
                        f"{name}: shape={x.shape}, sharding:",
                        y,
                        flush=True,
                    )

            jax.debug.inspect_array_sharding(x, callback=_cb)

    if isinstance(tree, jax.Array):
        _inspect(name, tree)
    else:
        jax.tree.map_with_path(
            lambda path, x: _inspect(name + jax.tree_util.keystr(path), x), tree
        )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from functools import partial

import jax
from jax import numpy as jnp

from netket.stats import Stats, statistics
from netket.utils.types import PyTree
from netket.utils.dispatch import dispatch
from netket.utils import HashablePartial
from netket.operator import (
    AbstractOperator,
    DiscreteOperator,
    Squared,
    ContinuousOperator,
    DiscreteJaxOperator,
)
from netket.operator._sum import SumGenericOperator
from netket.vqs.mc import (
    kernels,
    check_hilbert,
    get_local_kernel_arguments,
    get_local_kernel,
)

from .state import MCState


@dispatch
def get_local_kernel_arguments(vstate: MCState, Ô: Squared):  # noqa: F811
    check_hilbert(vstate.hilbert, Ô.hilbert)

    σ = vstate.samples
    σp, mels = Ô.parent.get_conn_padded(σ)
    return σ, (σp, mels)


@dispatch
def get_local_kernel(vstate: MCState, Ô: Squared):  # noqa: F811
    return kernels.local_value_squared_kernel


@dispatch
def get_local_kernel_arguments(vstate: MCState, Ô: DiscreteOperator):  # noqa: F811
    check_hilbert(vstate.hilbert, Ô.hilbert)

    σ = vstate.samples
    σp, mels = Ô.get_conn_padded(σ)
    return σ, (σp, mels)


@dispatch
def get_local_kernel(vstate: MCState, Ô: DiscreteOperator):  # noqa: F811
    return kernels.local_value_kernel


@dispatch
def get_local_kernel_arguments(vstate: MCState, Ô: DiscreteJaxOperator):  # noqa: F811
    check_hilbert(vstate.hilbert, Ô.hilbert)

    σ = vstate.samples
    return σ, Ô


@dispatch
def get_local_kernel(vstate: MCState, Ô: DiscreteJaxOperator):  # noqa: F811
    return kernels.local_value_kernel_jax


@dispatch
def get_local_kernel_arguments(vstate: MCState, Ô: ContinuousOperator):  # noqa: F811
    check_hilbert(vstate.hilbert, Ô.hilbert)
    return vstate.samples, Ô


@dispatch
def get_local_kernel(vstate: MCState, Ô: ContinuousOperator):  # noqa: F811
    # TODO: this should be moved other to dispatch in order to support MCMixedState
    return HashablePartial(
        lambda logpsi, params, x, op: op._expect_kernel(logpsi, params, x)
    )


## sum operators


@partial(get_local_kernel_arguments.dispatch, precedence=-2)
def get_local_kernel_args_sum(vs: MCState, O: SumGenericOperator):
    samples = vs.samples
    args = [get_local_kernel_arguments(vs, o)[1] for o in O.operators]
    return samples, (O.coefficients, args)


@partial(get_local_kernel.dispatch, precedence=-2)
def get_local_kernel_sum(vs: MCState, O: SumGenericOperator):
    O_kernels = tuple(get_local_kernel(vs, o) for o in O.operators)

    def sum_eval(logpsi, pars, σ, args, O_kernels):
        coeffs, O_args = args
        O_locs = 0
        for O_kernel, O_arg, k in zip(O_kernels, O_args, coeffs):
            O_locs = O_locs + k * O_kernel(logpsi, pars, σ, O_arg)
        return O_locs

    return HashablePartial(sum_eval, O_kernels=O_kernels)


# Standard implementation of expect for an MCState (pure) and a generic operator
# The dispatch rule is not strictly needed, as everything currently implemented
# in NetKet only defines a custom get_local_kernel_arguments and get_local_kernel
# but if somebody wants to override behaviour for an existing operator or define
# a completely arbitrary novel type of operator, this makes it much easier.
@dispatch
def expect(
    vstate: MCState, Ô: AbstractOperator, chunk_size: None
) -> Stats:  # noqa: F811
    σ, args = get_local_kernel_arguments(vstate, Ô)
    local_estimator_fun = get_local_kernel(vstate, Ô)

    return _expect(
        local_estimator_fun,
        vstate._apply_fun,
        vstate.sampler.machine_pow,
        vstate.parameters,
        vstate.model_state,
        σ,
        args,
    )


@partial(jax.jit, static_argnums=(0, 1))
def _expect(
    local_value_kernel: Callable,
    model_apply_fun: Callable,
    machine_pow: int,
    parameters: PyTree,
    model_state: PyTree,
    σ: jnp.ndarray,
    local_value_args: PyTree,
) -> Stats:
    n_chains = σ.shape[0]
    if σ.ndim >= 3:
        σ = jax.lax.collapse(σ, 0, 2)

    def logpsi(w, σ):
        return model_apply_fun({"params": w, **model_state}, σ)

    def log_pdf(w, σ):
        return machine_pow * model_apply_fun({"params": w, **model_state}, σ).real

    L_σ = local_value_kernel(logpsi, parameters, σ, local_value_args)
    Ō_stats = statistics(L_σ.reshape((n_chains, -1)))

    return Ō_stats



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netket.utils.config_flags import config

from netket.utils.moduletools import (
    _hide_submodules,
    rename_class,
    auto_export as _auto_export,
)
from netket.utils.version_check import module_version

# error if old dependencies are detected
from netket.utils import _dependencies_check

from netket.utils import dispatch
from netket.utils import struct
from netket.utils import numbers
from netket.utils import types
from netket.utils import float
from netket.utils import optional_deps
from netket.utils import timing
from netket.utils import display

from netket.utils.array import HashableArray, array_in
from netket.utils.partial import HashablePartial
from netket.utils.jax import get_afun_if_module, wrap_afun, wrap_to_support_scalar
from netket.utils.seed import random_seed
from netket.utils.summation import KahanSum

from netket.utils.holomorphic import is_probably_holomorphic

from netket.utils.deprecation import (
    warn_deprecation,
    deprecated,
    deprecated_new_name,
)

from netket.utils.model_frameworks import maybe_wrap_module

from netket.utils.history import History, accum_in_tree, accum_histories_in_tree

from netket.utils.static_range import StaticRange

# TODO: remove this import as it is deprecated
from netket.utils import mpi
from netket.utils import citations

_hide_submodules(
    __name__,
    remove_self=False,
    ignore=[
        "numbers",
        "types",
        "float",
        "dispatch",
        "errors",
        "timing",
        "display",
        "citations",
    ],
)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
3D axial symmetry groups and transformations.

This module provides functions to construct point groups and symmetry operations
for 3D systems with axial symmetry, including rotations, reflections, inversions,
and their combinations. It includes common point groups like cyclic (C), dihedral (D),
pyramidal (Cv), prismatic (Dh), and antiprismatic (Dd) groups.
"""

from itertools import count

import numpy as np

from netket.utils.float import is_approx_int
from netket.utils.types import Array
from netket.utils.moduletools import export, hide_unexported

from netket.utils.group._point_group import PGSymmetry, PointGroup
from netket.utils.group._semigroup import Identity

hide_unexported(__name__)

__all__ = ["rotation_group", "pyramidal", "dihedral", "prismatic", "antiprismatic"]


@export
def rotation(angle: float, axis: Array) -> PGSymmetry:
    """Returns a rotation by `angle` degrees around `axis`."""
    angle = np.radians(angle)
    axis = np.asarray(axis) / np.linalg.norm(axis)
    return PGSymmetry(
        np.cos(angle) * np.eye(3)
        + np.sin(angle) * np.cross(np.eye(3), axis)
        + (1 - np.cos(angle)) * np.outer(axis, axis)
    )


@export
def C(n: int, axis: Array = (0, 0, 1)) -> PointGroup:
    """
    Returns the :math:`C_n` `PointGroup` of rotations around a given axis.

    Arguments:
        n: the index of the rotation group (the smallest rotation angle is 360°/n)
        axis: the axis of rotations, need not be normalised, defaults to the z-axis

    Returns:
        a `PointGroup` implementing :math:`C_n`
    """
    return PointGroup(
        [Identity()] + [rotation(360 / n * i, axis) for i in range(1, n)], ndim=3
    )


rotation_group = C


@export
def screw(angle: float, trans: Array, origin: Array = (0, 0, 0)) -> PGSymmetry:
    """Returns a screw composed of a translation by `trans` and a rotation by `angle`
    degrees around its direction. The axis passes through `origin` (defaults to
    the Cartesian origin)."""
    angle = np.radians(angle)
    axis = np.asarray(trans) / np.linalg.norm(trans)
    W = (
        np.cos(angle) * np.eye(3)
        + np.sin(angle) * np.cross(np.eye(3), axis)
        + (1 - np.cos(angle)) * np.outer(axis, axis)
    )
    w = np.asarray(trans) + (np.eye(3) - W) @ np.asarray(origin)
    return PGSymmetry(W, w)


@export
def screw_group(angle: float, trans: Array, origin: Array = (0, 0, 0)) -> PointGroup:
    """Returns the `PointGroup` generated by a screw composed of a translation
    by `trans` and a rotation by `angle` degrees around its direction.

    The axis passes through `origin` (defaults to the Cartesian origin).
    The order of the group is controlled by `angle`.
    The output is only a valid `PointGroup` after supplying a `unit_cell`
    consistent with the screw axis; otherwise, operations like `product_table`
    will fail.
    """
    out = [Identity()]
    trans = np.asarray(trans)
    for i in count(start=1):
        if is_approx_int(i * angle / 360):
            break
        out.append(screw(i * angle, i * trans, origin))
    return PointGroup(out, ndim=3)


@export
def inversion() -> PGSymmetry:
    """Returns a 3D inversion as a `PGSymmetry`."""
    return PGSymmetry(-np.eye(3))


@export
def inversion_group() -> PGSymmetry:
    r"""
    :math:`\mathbb{Z}_2` `PointGroup` containing the identity and inversion across
    the origin.
    """
    return PointGroup([Identity(), inversion()], ndim=3)


@export
def reflection(axis: Array) -> PGSymmetry:
    r"""Returns a 3D reflection across a plane whose normal is `axis`.

    .. warning::

        For 2D reflections see
        :func:`netket.utils.group.planar.reflection`.

    Args:
        axis: The 3-component basis vector identifying the reflection axis,
            for example :code:`np.array([1,0,0])`.

    """
    if len(axis) != 3:
        raise ValueError(
            "The axis must be a 3 component vector."
            "If you want reflections in 2 dimensions, use "
            "nk.utils.group.planar.reflection instead."
        )

    axis = np.asarray(axis) / np.linalg.norm(axis)
    return PGSymmetry(np.eye(3) - 2 * np.outer(axis, axis))


@export
def reflection_group(axis: Array) -> PointGroup:
    r"""
    Returns the :math:`\mathbb{Z}_2` `PointGroup` containing the identity and a
    reflection across a plane with normal `axis`
    """
    return PointGroup([Identity(), reflection(axis)], ndim=3)


@export
def glide(axis: Array, trans: Array, origin: Array = (0, 0, 0)) -> PGSymmetry:
    r"""
    Returns a glide composed of translation by `trans` and a reflection across a
    plane that passes through `origin` and is normal to `axis`.

    `trans` and `axis` must be perpendicular.
    `origin` defaults to the Cartesian origin.
    """
    assert np.isclose(np.dot(axis, trans), 0.0)
    axis = np.asarray(axis) / np.linalg.norm(axis)
    W = np.eye(3) - 2 * np.outer(axis, axis)
    w = np.asarray(trans) + (np.eye(3) - W) @ np.asarray(origin)
    return PGSymmetry(W, w)


@export
def glide_group(axis: Array, trans: Array, origin: Array = (0, 0, 0)) -> PointGroup:
    r"""
    Returns the Z_2 `PointGroup` containing the identity and a glide composed of
    translation by `trans` and a reflection across a plane that passes through
    `origin` and is normal to `axis`.

    `trans` and `axis` must be perpendicular.
    `origin` defaults to the Cartesian origin.
    The output is only a valid `PointGroup` after supplying a `unit_cell`
    consistent with the glide plane; otherwise, operations like `product_table`
    will fail.
    """
    return PointGroup([Identity(), glide(axis, trans, origin)], ndim=3)


@export
def Ch(n: int, axis: Array = (0, 0, 1)) -> PointGroup:
    r"""
    Returns the reflection group :math:`C_{nh}` generated by an *n*-fold rotation axis
    and a reflection across the plane normal to the same axis.

    Arguments:
        n: index of the group
        axis: the axis of rotations and normal to the mirror plane, need not be
            normalised, defaults to the z-axis

    Returns:
        a `PointGroup` object implementing :math:`C_{nh}`
    """
    return reflection_group(axis) @ C(n, axis)


@export
def Cv(n: int, axis: Array = (0, 0, 1), axis2=(1, 0, 0)) -> PointGroup:
    r"""
    Returns the pyramidal group :math:`C_{nv}` generated by an *n*-fold rotation axis
    and a reflection across a plane that contains the same axis.

    Arguments:
        n: index of the group
        axis: the axis of rotations, need not be normalised, defaults to the z-axis
        axis2: normal of the generating mirror plane, need not be normalised, must
            be perpendicular to `axis`, defaults to the x-axis

    Returns:
        a `PointGroup` object implementing :math:`C_{nv}`
    """
    assert np.isclose(np.dot(axis, axis2), 0.0)
    return reflection_group(axis2) @ C(n, axis)


pyramidal = Cv


@export
def rotoreflection(angle: float, axis: Array) -> PGSymmetry:
    """Returns a rotoreflection by `angle` degrees around `axis`."""
    angle = np.radians(angle)
    axis = np.asarray(axis) / np.linalg.norm(axis)
    rot_matrix = (
        np.cos(angle) * np.eye(3)
        + np.sin(angle) * np.cross(np.eye(3), axis)
        + (1 - np.cos(angle)) * np.outer(axis, axis)
    )
    refl_matrix = np.eye(3) - 2 * np.outer(axis, axis)
    return PGSymmetry(refl_matrix @ rot_matrix)


@export
def S(n: int, axis: Array = (0, 0, 1)) -> PointGroup:
    """
    Returns the :math:`S_n` `PointGroup` of rotoreflections around a given axis.

    Arguments:
        n: the index of the rotoreflection group (the smallest rotation angle is 360°/n)
        axis: the axis, need not be normalised, defaults to the z-axis

    Returns:
        a `PointGroup` implementing :math:`S_n`
    """
    if n % 2 == 1:
        return Ch(n, axis)
    else:
        return PointGroup(
            [Identity()]
            + [
                (
                    rotoreflection(360 / n * i, axis)
                    if i % 2 == 1
                    else rotation(360 / n * i, axis)
                )
                for i in range(1, n)
            ],
            ndim=3,
        )


rotoreflection_group = S
export(rotoreflection_group)


@export
def D(n: int, axis: Array = (0, 0, 1), axis2: Array = (1, 0, 0)) -> PointGroup:
    """
    Returns the dihedral group :math:`D_n` generated by an *n*-fold rotation axis
    and a twofold rotation axis perpendicular to the former.

    Arguments:
        n: index of the group
        axis: the n-fold rotation axis, need not be normalised, defaults to the z-axis
        axis2: generating twofold rotation axis, need not be normalised, must be
            perpendicular to `axis`, defaults to the x-axis

    Returns:
        a `PointGroup` object implementing :math:`D_n`
    """
    assert np.isclose(np.dot(axis, axis2), 0.0)
    return C(2, axis2) @ C(n, axis)


dihedral = D


@export
def cuboid_rotations() -> PointGroup:
    """Rotational symmetries of a cuboid with edges aligned with the Cartesian axes."""
    return D(2)


@export
def Dh(n: int, axis: Array = (0, 0, 1), axis2=(1, 0, 0)) -> PointGroup:
    """
    Returns the prismatic group :math:`D_{nh}` generated by an *n*-fold rotation axis,
    a twofold rotation axis perpendicular to the former, and a mirror plane normal
    to the *n*-fold axis.

    Arguments:
        n: index of the group
        axis: the n-fold rotation axis and normal to the mirror plane,
            need not be normalised, defaults to the z-axis
        axis2: generating twofold rotation axis, need not be normalised,
            must be perpendicular to `axis`, defaults to the x-axis

    Returns:
        a `PointGroup` object implementing :math:`D_{nh}`
    """
    return reflection_group(axis) @ D(n, axis, axis2)


prismatic = Dh


@export
def cuboid() -> PointGroup:
    """Symmetry group of a cuboid with edges aligned with the Cartesian axes."""
    return Dh(2)


@export
def Dd(n: int, axis: Array = (0, 0, 1), axis2=(1, 0, 0)) -> PointGroup:
    """
    Returns the antiprismatic group :math:`D_{nd}` generated by a 2*n*-fold roto-
    reflection axis and a reflection across a plane that contains the same axis.

    Arguments:
        n: index of the group
        axis: the rotoreflection axis, need not be normalised, defaults to the z-axis
        axis2: normal of the generating mirror plane, need not be normalised, must be
            perpendicular to `axis`, defaults to the x-axis

    Returns:
        a `PointGroup` object implementing :math:`D_{nd}`
    """
    assert np.isclose(np.dot(axis, axis2), 0.0)
    return reflection_group(axis2) @ S(2 * n, axis)


antiprismatic = Dd



import inspect

from collections.abc import Callable
from functools import partial


def ensure_accepts_kwargs(f: Callable, kwarg_name: str):
    """
    Add a keyword argument to a function if it is not already present.

    This is useful to make a function compatible with a certain interface, even if it does not
    need the additional argument.

    In this case, we use it to make the linear solver functions compatible with the QGT interface.
    Some QGT solvers require the vector of local energies as an additional argument. The standard
    solver interface however only requires the matrix and the right-hand side (jacobian @ local_energies).

    Args:
        f: The function to concretize.
        kwarg_name: The name of the keyword argument to add.

    Returns:
        A new function that accepts the keyword argument `kwarg_name`.
    """
    sig = inspect.signature(f)
    params = sig.parameters

    if kwarg_name in params:
        return f
    else:
        to_concretize = f
        if isinstance(f, partial):
            to_concretize = f.func

        def concretized_func(*args, **kwargs):
            kwargs.pop(kwarg_name, None)
            return to_concretize(*args, **kwargs)

        if isinstance(f, partial):
            concretized_func = partial(concretized_func, *f.args, **f.keywords)

        return concretized_func



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from math import prod

import numpy as np
import jax.numpy as jnp
from jax.tree_util import register_pytree_node_class

from netket.operator._discrete_operator_jax import DiscreteJaxOperator

from netket.operator._prod.base import ProductOperator


@register_pytree_node_class
class ProductDiscreteJaxOperator(ProductOperator, DiscreteJaxOperator):
    def __init__(self, *operators, coefficient=1.0, dtype=None):
        if not all(isinstance(op, DiscreteJaxOperator) for op in operators):
            raise TypeError(
                "Arguments to ProductDiscreteJaxOperator must all be "
                "subtypes of DiscreteJaxOperator. However the types are:\n\n"
                f"{list(type(op) for op in operators)}\n"
            )
        self._initialized = False
        super().__init__(
            operators, operators[0].hilbert, coefficient=coefficient, dtype=dtype
        )

    @property
    def max_conn_size(self) -> int:
        """The maximum number of non zero ⟨x|O|x'⟩ for every x."""
        return prod(op.max_conn_size for op in self.operators)

    def get_conn_padded(self, x: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        x_r = x.reshape(-1, 1, x.shape[-1])
        Ns, N = x_r.shape[0], x_r.shape[-1]
        samples = x
        mels = jnp.full((1, 1, 1), self.coefficient)
        for op in self.operators:
            op_ys, op_mels = op.get_conn_padded(samples)
            samples = op_ys.reshape(Ns, -1, N)
            mels = (mels * op_mels).reshape(Ns, -1, 1)

        ys = samples.reshape(*x.shape[:-1], -1, x.shape[-1])
        mels = mels.reshape(*ys.shape[:-1])
        return ys, mels

    def to_numba_operator(self) -> "ProductOperator":  # noqa: F821
        """
        Returns the standard (numba) version of this operator, which is an
        instance of {class}`nk.operator.Ising`.
        """
        from netket.operator._prod.discrete_operator import ProductDiscreteOperator

        ops_numba = tuple(op.to_numba_operator() for op in self.operators)
        return ProductDiscreteOperator(
            *ops_numba, coefficients=self.coefficients, dtype=self.dtype
        )

    def tree_flatten(self):
        data = (self.operators, self.coefficient)
        metadata = {"dtype": self.dtype}
        return data, metadata

    @classmethod
    def tree_unflatten(cls, metadata, data):
        ops, coefficient = data
        dtype = metadata["dtype"]

        # Create instance without calling __init__ to avoid ArgInfo validation issues
        obj = object.__new__(cls)
        obj._operators = tuple(ops)
        obj._coefficient = coefficient
        obj._dtype = dtype
        obj._initialized = False

        # Set the hilbert space from the first operator
        obj._hilbert = ops[0].hilbert

        return obj



# Copyright 2023-2024 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# this file contains a more-or-less 1:1 port of the numba localoperator to jax, using padding where necessary

from functools import partial
from typing import TYPE_CHECKING

import jax
import jax.numpy as jnp
from jax.tree_util import register_pytree_node_class

from netket.errors import JaxOperatorNotConvertibleToNumba

from .base import LocalOperatorBase
from .compile_helpers import pack_internals_jax

from .._pauli_strings import PauliStringsJax
from .._discrete_operator_jax import DiscreteJaxOperator

if TYPE_CHECKING:
    from .numba import LocalOperatorNumba


@partial(jax.vmap, in_axes=(0, None))  # samples
@partial(jax.vmap, in_axes=(0, 0))  # operators
def _state_to_number(x_i, basis_i):
    # convert array of local states to number
    # in the hilbert space of all the sites the operator is acting on

    # number of sites these operators are acting on
    acting_size_i = basis_i.shape[-1]

    @partial(jax.vmap, in_axes=(None, None, None, 0))  # vmap over local sites (k)
    def _f(acting_size_i, x_i, basis_i, k):
        # special case for empty operator
        if len(basis_i) == 0:
            return jnp.zeros((), dtype=basis_i.dtype)
        return basis_i[k] * x_i[acting_size_i - k - 1]

    # vmap over sites we are acting on, computing the contribution to the total int,
    # then taking the sum
    return _f(acting_size_i, x_i, basis_i, jnp.arange(acting_size_i)).sum()


@partial(jax.vmap, in_axes=(0, 0, None))  # Ns
@partial(jax.vmap, in_axes=(None, 0, 0))  # rows
@partial(jax.vmap, in_axes=(None, 0, None))  # ncmax
def _set_at(x, new_x_ao, acting_on):
    return x.at[acting_on].set(new_x_ao)


@partial(jax.vmap, in_axes=(None, 0))  # Ns
@partial(jax.vmap, in_axes=(0, 0))  # rows
def _index_at(diag_mels, i):
    # special case for empty operator
    if len(diag_mels) == 0:
        return jnp.zeros((), dtype=diag_mels.dtype)
    return diag_mels[i]


@partial(jax.jit, static_argnums=(0, 1))
def _local_operator_kernel_jax(nonzero_diagonal, max_conn_size, mel_cutoff, op_args, x):
    assert x.ndim == 2

    # in the foollowing variables with a trailing underscore are per group
    # of operators with the same number of sites they act on

    (
        acting_on_,
        n_conns_,
        diag_mels_,
        x_prime_,
        mels_,
        basis_,
        constant,
    ) = op_args

    n_groups = len(acting_on_)
    ncmax_ = [m.shape[2] for m in mels_]

    ###
    # determine the row x corresponds to for each of the operators
    #
    i_row_ = []
    # iterate over all groups of operators acting on a certain number of sites
    for k in range(n_groups):
        # extract the local states of all sites the operators are acting on
        # and compute the corresponding number / row index
        i_row_.append(_state_to_number(x[:, acting_on_[k]], basis_[k]))
    ###
    # compute the number of connected elements
    #
    # extract the number of (nonzero) off-diagonal elements of the rows
    n_conn_offdiag_ = tuple(map(_index_at, n_conns_, i_row_))
    # sum for each group of operators acting on a certain number of sites
    # and sum over all gropus to get the total
    n_conn_offdiag = sum([n.sum(axis=-1) for n in n_conn_offdiag_])
    n_conn_diag = 1 if nonzero_diagonal else 0
    n_conn_total = n_conn_diag + n_conn_offdiag

    mels_offdiag_ = []
    xp_offdiag_ = []

    if nonzero_diagonal:
        xp_diag = x
        # extract diagonal mels of the rows
        mels_diag_ = tuple(map(_index_at, diag_mels_, i_row_))
        # sum over operators
        mels_diag = constant + sum([m.sum(axis=-1) for m in mels_diag_])
    else:
        # zero diagonal
        mels_diag = None
        xp_diag = None

    if mels_diag is not None:
        mels_offdiag_.append(mels_diag)
        xp_offdiag_.append(xp_diag)

    # iterate over all groups of operators acting on a certain number of sites
    for k in range(n_groups):
        # indices used to select rows
        i = i_row_[k]
        a = jnp.arange(i.shape[1])

        # mask for the connected elements
        # (not all operators have the same number nonzeros per row, and the
        # arrays we use here are padded to the max within each group)
        # this mask is False whenever it's just padding
        conn_maskall = jnp.arange(ncmax_[k])[None, None] < n_conns_[k][a, i][:, :, None]

        # set mels in the padding to 0
        # TODO check this is necessary (shouldn't it already be 0 ???)
        mels_offdiag = mels_[k][a, i] * conn_maskall
        #
        # compute xp
        #
        # we start from the xp for the sites the operator is acting on
        new = x_prime_[k][a, i].astype(x.dtype)  # (Ns, terms, ncmax, nsitesactiongon)
        acting_on = acting_on_[k]
        old = x[:, acting_on]  # (Ns, terms, nsitesactiongon)
        old = jnp.broadcast_to(old[:, :, None, :], new.shape)
        mask = jnp.broadcast_to(conn_maskall[:, :, :, None], new.shape)
        # select it only if we are not padding, otherwise keep old
        new_x_ao = jax.lax.select(mask, new, old)
        # now insert the local states into the full x
        xp_offdiag = _set_at(x, new_x_ao, acting_on)

        mels_offdiag_.append(mels_offdiag)
        xp_offdiag_.append(xp_offdiag)

    if max_conn_size is not None:
        # compute a mask which tells us where the actual mels are and
        # where there is just padding
        # (we could also just be lazy and check the mels for being 0,
        # but allows us to make the order of mels consistent with the numba op)
        mask_ = []
        if mels_diag is not None:
            mask_.append(jnp.full(mels_diag.shape, fill_value=True))
        # iterate over all groups of operators acting on a certain number of sites
        for k in range(n_groups):
            mask = jnp.arange(ncmax_[k])[None, None, :] < n_conn_offdiag_[k][:, :, None]
            mask_.append(mask)

        # pad with old state and mel 0
        xp_offdiag_.append(x[:, None][:, :1])
        mels_offdiag_.append(jnp.zeros(x.shape[:-1]))
        mask_.append(jnp.full((x.shape[0], 1), fill_value=False))

    mels = jnp.hstack([m.reshape(m.shape[0], -1) for m in mels_offdiag_])
    xp = jnp.hstack([x.reshape(x.shape[0], -1, x.shape[-1]) for x in xp_offdiag_])

    # TODO run unique on it, there might be repeated xps

    if max_conn_size is None:
        return xp, mels, n_conn_total
    else:
        if mel_cutoff is not None:
            mask = jnp.abs(mels) > mel_cutoff
            n_conn_total = mask.sum(axis=-1)
        else:
            mask = jnp.hstack([m.reshape(m.shape[0], -1) for m in mask_])
        # move nonzero mels to the front and keep exactly max_conn_size
        (ind,) = jax.vmap(partial(jnp.where, size=max_conn_size, fill_value=-1))(mask)

        xp = jax.vmap(partial(jnp.take, axis=0))(xp, ind)
        mels = jax.vmap(partial(jnp.take, axis=0))(mels, ind)

        return xp, mels, n_conn_total


@register_pytree_node_class
class LocalOperatorJax(LocalOperatorBase, DiscreteJaxOperator):
    """Jax implementation of an operator composed of a sum of
    local terms, each of which acts on a small number of sites.

    .. warning::

        The complexity of the constructor of this operator scales
        exponentially with the number of sites on which each term
        acts, and linearly with the number of terms.

        Refrain from using this for terms acting on more than
        6-sites. For Spin-1/2 systems, prefer instead
        :class:`netket.operator.PauliStrings`. For non spin-1/2
        there is nothing that will work efficiently out of the box,
        but you can easily roll your own.

    .. warning::

        The runtime complexity is proportional to the number of
        connected entries
        :attr:`~netket.operator.DiscreteJaxOperator.max_conn_size`.

    For the numba-based implementation, look at
    :class:`netket.operator.LocalOperatorNumba`.
    """

    _convertible: bool = True
    """Internal flag. True if it can be converted to numba, false
    otherwise."""

    def _setup(self, force=False):
        if force or not self._initialized:
            data = pack_internals_jax(
                self.hilbert,
                self._operators_dict,
                self.constant,
                self.dtype,
                self.mel_cutoff,
            )

            self._acting_on = data["acting_on"]
            self._n_conns = data["n_conns"]
            self._diag_mels = data["diag_mels"]
            self._x_prime = data["x_prime"]
            self._mels = data["mels"]
            self._basis = data["basis"]
            self._nonzero_diagonal = data["nonzero_diagonal"]
            self._max_conn_size = data["max_conn_size"]

            self._initialized = True

    def _get_conn_padded(self, x):
        self._setup()

        shape = x.shape
        x = x.reshape(-1, x.shape[-1])
        x_ids = self.hilbert.states_to_local_indices(x)

        xp_ids, mels, n_conn = _local_operator_kernel_jax(
            self._nonzero_diagonal,
            self._max_conn_size,
            self._mel_cutoff,
            (
                self._acting_on,
                self._n_conns,
                self._diag_mels,
                self._x_prime,
                self._mels,
                self._basis,
                self._constant,
            ),
            x_ids,
        )

        xp = self.hilbert.local_indices_to_states(xp_ids, dtype=x.dtype)
        xp = xp.reshape(shape[:-1] + xp.shape[-2:])
        mels = mels.reshape(shape[:-1] + mels.shape[-1:])
        n_conn = n_conn.reshape(shape[:-1])

        return xp, mels, n_conn

    def get_conn_padded(self, x):
        xp, mels, _ = self._get_conn_padded(x)
        return xp, mels

    def n_conn(self, x, out=None):
        if out is not None:
            raise NotImplementedError()
        _, _, n_conn = self._get_conn_padded(x)
        return n_conn

    def tree_flatten(self):
        self._setup()
        data = (
            self._acting_on,
            self._n_conns,
            self._diag_mels,
            self._x_prime,
            self._mels,
            self._basis,
            self._constant,
        )
        metadata = {
            "hilbert": self.hilbert,
            "dtype": self.dtype,
            "nonzero_diagonal": self._nonzero_diagonal,
            "max_conn_size": self._max_conn_size,
        }
        return data, metadata

    @classmethod
    def tree_unflatten(cls, metadata, data):
        hi = metadata["hilbert"]
        dtype = metadata["dtype"]

        op = cls(hi, dtype=dtype)

        op._nonzero_diagonal = metadata["nonzero_diagonal"]
        op._max_conn_size = metadata["max_conn_size"]
        (
            op._acting_on,
            op._n_conns,
            op._diag_mels,
            op._x_prime,
            op._mels,
            op._basis,
            op._constant,
        ) = data

        op._initialized = True
        op._convertible = False
        return op

    def to_numba_operator(self) -> "LocalOperatorNumba":  # noqa: F821
        """
        Returns the standard numba version of this operator, which is an
        instance of :class:`netket.operator.LocalOperator`.
        """
        from .numba import LocalOperatorNumba

        if self._convertible is False:
            raise JaxOperatorNotConvertibleToNumba(self)

        return LocalOperatorNumba(
            self.hilbert,
            self.operators,
            self.acting_on,
            self.constant,
            dtype=self.dtype,
            mel_cutoff=self.mel_cutoff,
        )

    def to_pauli_strings(self) -> "PauliStringsJax":  # noqa: F821
        """Convert to PauliStrings object"""
        return super().to_pauli_strings(
            cls=PauliStringsJax,
        )



"""
Sphinx extension for NetKet citation system.

This extension provides directives to display citation information from NetKet's
citation system.
"""

from docutils import nodes
from sphinx.application import Sphinx
from sphinx.util.docutils import SphinxDirective
from sphinx.util import logging


logger = logging.getLogger(__name__)


def _extract_doi_from_bibtex(bibtex_entry):
    """Extract DOI from bibtex entry."""
    import re

    # Pattern to match doi field
    doi_patterns = [
        r'doi\s*=\s*{([^}]+)}',
        r'doi\s*=\s*"([^"]+)"',
        r'doi\s*=\s*([^,\s}]+)',
    ]

    for pattern in doi_patterns:
        match = re.search(pattern, bibtex_entry, re.IGNORECASE)
        if match:
            return match.group(1).strip()

    return None


def _extract_url_from_bibtex(bibtex_entry):
    """Extract URL from bibtex entry."""
    import re

    # Pattern to match url field
    url_patterns = [
        r'url\s*=\s*{([^}]+)}',
        r'url\s*=\s*"([^"]+)"',
        r'url\s*=\s*([^,\s}]+)',
    ]

    for pattern in url_patterns:
        match = re.search(pattern, bibtex_entry, re.IGNORECASE)
        if match:
            return match.group(1).strip()

    return None


def _extract_title_from_bibtex(bibtex_entry):
    """Extract and clean title from bibtex entry."""
    import re

    # Pattern to match title field
    title_patterns = [
        r'title\s*=\s*{([^}]+)}',
        r'title\s*=\s*"([^"]+)"',
    ]

    for pattern in title_patterns:
        match = re.search(pattern, bibtex_entry, re.IGNORECASE)
        if match:
            title = match.group(1).strip()
            # Remove LaTeX formatting
            title = re.sub(r'\\[a-zA-Z]+{([^}]*)}', r'\1', title)
            title = re.sub(r'[{}]', '', title)
            return title

    return None


def _generate_bootstrap_citations():
    """Generate bootstrap card-based citation HTML."""
    from netket.utils.citations import get_all_citations

    # Get all citations
    citation_registry = get_all_citations()

    if not citation_registry:
        return '<div class="alert alert-info">No citations have been registered.</div>'

    # Filter out NetKet papers
    netket_papers = ["netket3:2022", "netket2:2019", "netket3:2021"]
    filtered_registry = {k: v for k, v in citation_registry.items() if k not in netket_papers}

    if not filtered_registry:
        return '<div class="alert alert-info">No algorithm-specific citations have been registered.</div>'

    # Group citations by group_id
    groups = {}
    for key, entry in filtered_registry.items():
        group_id = entry.get('group_id')
        if group_id not in groups:
            groups[group_id] = []
        groups[group_id].append((key, entry))

    html_parts = ['<div class="row">']

    for group_id, group_entries in groups.items():
        # Get the first entry to extract group-level info
        first_key, first_entry = group_entries[0]
        condition = first_entry.get('condition', '').strip()
        message = first_entry.get('message', 'This work used the methods described in Ref.~').strip()

        # Create bootstrap card
        html_parts.append('<div class="col-12 mb-2">')
        html_parts.append('<div class="card">')
        html_parts.append('<div class="card-body p-2">')

        # Combine condition, message and references in one paragraph
        refs_html = []
        for ref_key, entry in group_entries:
            bibtex = entry.get('bibtex', '')
            if bibtex:
                title = _extract_title_from_bibtex(bibtex)
                doi = _extract_doi_from_bibtex(bibtex)
                url = _extract_url_from_bibtex(bibtex)

                # Create link
                if doi:
                    link_url = f"https://doi.org/{doi}"
                elif url:
                    link_url = url
                else:
                    link_url = None

                if link_url and title:
                    refs_html.append(f'(<code>{ref_key}</code>)<a href="{link_url}" target="_blank">"{title}"</a>')
                elif title:
                    refs_html.append(f'(<code>{ref_key}</code>)"{title}"')
                else:
                    refs_html.append(f'<code>{ref_key}</code>')
            else:
                refs_html.append(f'<code>{ref_key}</code>')

        # Create single paragraph with condition (if exists), message, and references
        html_parts.append('<p class="card-text mb-1">')
        if condition:
            html_parts.append(f'<strong>{condition}:</strong> ')
        html_parts.append(f'<em>{message}</em>')
        if refs_html:
            html_parts.append(' <small class="text-muted">')
            html_parts.append(' '.join(refs_html))
            html_parts.append('</small>')
        html_parts.append('</p>')

        # Add collapsible BibTeX entries
        bibtex_entries = [entry.get('bibtex', '') for _, entry in group_entries if entry.get('bibtex')]
        if bibtex_entries:
            html_parts.append('<details>')
            html_parts.append('<summary><small>BibTeX entries</small></summary>')
            html_parts.append('<pre class="mb-0"><code class="language-bibtex">')
            html_parts.append('\n\n'.join(bibtex_entries))
            html_parts.append('</code></pre>')
            html_parts.append('</details>')

        html_parts.append('</div>')  # card-body
        html_parts.append('</div>')  # card
        html_parts.append('</div>')  # col

    html_parts.append('</div>')  # row

    return ''.join(html_parts)


class CitationsListNode(nodes.Element):
    """Custom node for citations list."""
    pass


class CitationsListDirective(SphinxDirective):
    """
    Directive to display all citations from NetKet's citation system.
    
    Example:
        .. netket-citations::
           :format: detailed
    """

    has_content = False
    optional_arguments = 0
    option_spec = {
        'format': str,  # 'detailed' or 'simple'
    }

    def run(self):
        node = CitationsListNode()
        node['format'] = self.options.get('format', 'detailed')
        return [node]


def visit_citations_list_node(self, node):
    """Visit function for CitationsListNode."""
    format_type = node['format']

    try:
        # Import NetKet to load citation registrations
        import netket as nk
        from netket.utils.citations import get_all_citations
        import re
        import uuid

        if format_type == 'detailed':
            citations_html = _generate_bootstrap_citations()
        else:
            from netket.utils.citations.docs import get_citations_for_docs
            citations_html = get_citations_for_docs(format_type)

        self.body.append(citations_html)

    except ImportError as e:
        logger.warning(f"Could not import NetKet for citations: {e}")
        self.body.append('<div class="admonition warning"><p class="admonition-title">Warning</p>')
        self.body.append('<p>Citations could not be loaded. NetKet import failed.</p></div>')
    except Exception as e:
        logger.warning(f"Error generating citations: {e}")
        self.body.append('<div class="admonition error"><p class="admonition-title">Error</p>')
        self.body.append(f'<p>Error generating citations: {e}</p></div>')


def depart_citations_list_node(self, node):
    """Depart function for CitationsListNode."""
    pass


def setup(app: Sphinx):
    """Setup function for the Sphinx extension."""

    # Add the custom directive
    app.add_directive('netket-citations', CitationsListDirective)

    # Add the custom node and its handlers
    app.add_node(
        CitationsListNode,
        html=(visit_citations_list_node, depart_citations_list_node),
        latex=(visit_citations_list_node, depart_citations_list_node),
        text=(visit_citations_list_node, depart_citations_list_node),
    )

    return {
        'version': '1.0',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
2D planar symmetry groups and transformations.

This module provides functions to construct point groups and symmetry operations
for 2D systems, including rotations, reflections, and glide operations.
It includes common 2D point groups like cyclic (C) and dihedral (D) groups,
as well as specific groups for rectangles and squares.
"""

import numpy as np

from netket.utils.types import Array
from netket.utils.moduletools import export, hide_unexported

from netket.utils.group._point_group import PGSymmetry, PointGroup
from netket.utils.group._semigroup import Identity


hide_unexported(__name__)

__all__ = [
    "rotation_group",
    "dihedral",
]


@export
def rotation(angle: float) -> PGSymmetry:
    """
    Returns a 2D rotation by `angle` degrees.

    Args:
       angle: the angle of rotation in degrees.
    """
    angle = np.radians(angle)
    return PGSymmetry(
        np.asarray([[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]])
    )


@export
def C(n: int) -> PointGroup:
    """
    Returns the Cₙ `PointGroup` of 2D rotational symmetries.

    Arguments:
        n: the index of the rotation group (the smallest rotation angle is 360°/n)

    Returns:
        a `PointGroup` implementing Cₙ
    """
    return PointGroup(
        [Identity()] + [rotation(360 / n * i) for i in range(1, n)], ndim=2
    )


rotation_group = C


@export
def reflection(angle: float) -> PGSymmetry:
    """
    Returns a 2D reflection across an axis at angle `angle` to the +x direction.

    Args:
        angle: the angle between the +x axis and the reflection axis.
    """
    axis = np.radians(angle) * 2  # the mirror matrix is written in terms of 2φ
    return PGSymmetry(
        np.asarray([[np.cos(axis), np.sin(axis)], [np.sin(axis), -np.cos(axis)]])
    )


@export
def reflection_group(angle: float) -> PointGroup:
    """
    Returns the Z₂ `PointGroup` containing the identity and a reflection across an
    axis at angle `angle` to the +x.

    Args:
        angle: the angle between the +x axis and the reflection axis.
    """
    return PointGroup([Identity(), reflection(angle)], ndim=2)


@export
def glide(trans: Array, origin: Array = (0, 0)) -> PGSymmetry:
    """
    Returns a 2D glide composed of translation by `trans` and reflection across
    its direction.

    Arguments:
        trans: translation vector
        origin: a point on the glide axis, defaults to the origin
    """
    axis = np.arctan2(trans[1], trans[0]) * 2
    W = np.asarray([[np.cos(axis), np.sin(axis)], [np.sin(axis), -np.cos(axis)]])
    w = np.asarray(trans) + (np.eye(2) - W) @ np.asarray(origin)
    return PGSymmetry(W, w)


@export
def glide_group(trans: Array, origin: Array = (0, 0)) -> PGSymmetry:
    """
    Returns the Z_2 `PointGroup`containing the identity and a  2D glide composed
    of translation by `trans` and reflection across its direction.

    Arguments:
        trans: translation vector
        origin: a point on the glide axis, defaults to the origin
    The output is only a valid `PointGroup` after supplying a `unit_cell`
    consistent with the glide axis; otherwise, operations like `product_table`
    will fail.
    """
    return PointGroup([Identity(), glide(trans, origin)], ndim=2)


@export
def D(n: int, axis: float = 0) -> PointGroup:
    """
    Returns the 2D dihedral `PointGroup` :math:`D_n` generated by a 360°/n rotation
    and a reflection.

    Arguments:
        n: index of the dihedral group
        axis: optional, the angle of one reflection axis with the +x direction
            (default: 0)

    Returns:
        a `PointGroup` object implementing :math:`D_n`
    """
    return reflection_group(axis) @ C(n)


dihedral = D


@export
def rectangle() -> PointGroup:
    """
    The symmetry group of a rectangle aligned with the Cartesian axes (Vierergruppe).
    """
    return D(2)


@export
def square() -> PointGroup:
    """The symmetry group of a square aligned with the Cartesian axes."""
    return D(4)



import jax
import jax.numpy as jnp

from flax import nnx
from flax import linen

from netket.sampler import MetropolisSamplerState
from netket.jax import PRNGKey
from netket.vqs import MCState, FullSumState
from netket.operator import ContinuousOperator, DiscreteOperator
from netket.operator._prod.base import ProductOperator
from netket._src.nn.apply_operator.linen import ApplyOperatorModuleLinen
from netket._src.nn.apply_operator.nnx import ApplyOperatorModuleNNX
from netket._src.nn.apply_operator.functional import make_logpsi_op_afun


def apply_operator(operator, vstate, *, seed=None):
    """
    Apply an operator to a variational state.

    The returned variational state wraps the model of vstate with an operator transformation.
    The implementation depends on the model type:

    - **Linen modules**: Wrapped in an ApplyOperatorModuleLinen. Operator stored in flattened form with
      leaves accessible as `op_vstate.variables['operator']['leaves']` and treedef stored in
      the module as `op_vstate._model.operator_treedef`.
    - **NNX modules**: Wrapped in an ApplyOperatorModuleNNX. Operator accessible as
      `op_vstate.model.operator`.
    - **Other (functional)**: Uses the functional approach with make_logpsi_op_afun.
      Operator accessible as `op_vstate.variables['operator']`.

    .. note::

        Note that is the vstate's chunk size is specified, the chunk size of the transformed vstate will
        be set to `vstate.chunk_size // operator.max_conn_size` to account for the increased memory usage.

    .. note::

        When applying an operator to a vstate that already has an operator applied (nested application),
        the operators are automatically combined into a ProductOperator to avoid double wrapping.
        For example, B@[A@psi] will be computed as (B*A)@psi instead of B@(A@psi).

    Args:
        operator: The operator to apply in front of the variational state ket
        vstate: The variational state (or ket)
    """

    if not isinstance(vstate, (FullSumState, MCState)):
        raise TypeError("vstate must be either an MCState or a FullSumState.")

    base_module = vstate.model
    base_variables = vstate.variables

    # sometimes we apply a joint operator...
    applied_operator = operator

    if isinstance(base_module, linen.Module):
        if isinstance(vstate.model, ApplyOperatorModuleLinen):
            existing_operator = jax.tree.unflatten(
                vstate.model.operator_treedef, vstate.variables["operator"]["leaves"]
            )
            base_module = vstate.model.base_module
            base_variables = {
                collection: params.get("base_module", params)
                for collection, params in vstate.variables.items()
                if collection != "operator"
            }
            applied_operator = ProductOperator(operator, existing_operator)

        transformed_module, new_variables = (
            ApplyOperatorModuleLinen.from_module_and_variables(
                base_module, applied_operator, base_variables
            )
        )
        transformed_apply_fun = None

    elif isinstance(base_module, nnx.Module):
        if isinstance(vstate.model, ApplyOperatorModuleNNX):
            existing_operator = vstate.model.operator
            base_module = vstate.model.base_module
            applied_operator = ProductOperator(operator, existing_operator)

        transformed_module = ApplyOperatorModuleNNX(base_module, applied_operator)
        new_variables = None
        transformed_apply_fun = None

    else:
        if "operator" in base_variables:
            raise NotImplementedError(
                "Applying an operator to a variational state that already has an operator applied "
                "is not supported for functional (apply_fun) variational states. "
                "This feature is only supported for Linen and NNX modules."
            )
        transformed_apply_fun, new_variables = make_logpsi_op_afun(
            vstate._apply_fun, applied_operator, base_variables
        )
        transformed_module = None

    if vstate.chunk_size is None:
        chunk_size = None
    else:
        chunk_size = max(vstate.chunk_size // operator.max_conn_size, 1)

    if isinstance(vstate, FullSumState):
        del seed

        return FullSumState(
            hilbert=vstate.hilbert,
            model=transformed_module,
            apply_fun=transformed_apply_fun,
            variables=new_variables,
            chunk_size=chunk_size,
        )

    elif isinstance(vstate, MCState):
        # Module-based approach (Linen or NNX)
        transformed_vstate = MCState(
            sampler=vstate.sampler,
            model=transformed_module,
            apply_fun=transformed_apply_fun,
            variables=new_variables,
            n_samples=vstate.n_samples,
            n_discard_per_chain=vstate.n_discard_per_chain,
            chunk_size=chunk_size,
            sampler_seed=seed,
        )
        if isinstance(vstate.sampler_state, MetropolisSamplerState):
            if isinstance(operator, DiscreteOperator):
                # fold in a random counter so that we don't have the same seed as sampler
                # itself.
                seed = jax.random.fold_in(PRNGKey(seed), 123)

                x = vstate.sampler_state.σ
                xp, mels = operator.get_conn_padded(x)
                ids = jax.random.randint(seed, (x.shape[0],), 0, operator.max_conn_size)
                new_x = xp[jnp.arange(x.shape[0]), ids, :]
                transformed_vstate.sampler_state = (
                    transformed_vstate.sampler_state.replace(σ=new_x)
                )
            elif isinstance(operator, ContinuousOperator):
                # TODO: Implement
                pass

        return transformed_vstate



from functools import partial

import jax
import jax.numpy as jnp

from netket.vqs import MCState, expect
from netket.stats import Stats

from netket.experimental.observable.infidelity.infidelity_operator import (
    InfidelityOperator,
)


@expect.dispatch
def infidelity(vstate: MCState, op: InfidelityOperator, chunk_size: None):
    if op.hilbert != vstate.hilbert:
        raise TypeError("Hilbert spaces should match")

    return infidelity_sampling_inner(
        vstate._apply_fun,
        op.target_state._apply_fun,
        vstate.parameters,
        op.target_state.parameters,
        vstate.model_state,
        op.target_state.model_state,
        vstate.samples,
        op.target_state.samples,
        op.cv_coeff,
    )


@partial(jax.jit, static_argnames=("afun", "afun_t"))
def get_kernels(afun, afun_t, params, params_t, σ, σ_t, model_state, model_state_t):
    W = {"params": params, **model_state}
    W_t = {"params": params_t, **model_state_t}

    log_val = afun_t(W_t, σ) - afun(W, σ)
    log_val_t = afun(W, σ_t) - afun_t(W_t, σ_t)

    return log_val, log_val_t


def get_local_estimator(
    vstate, target_state, samples, weights, samples_t, weights_t, cv_coeff=-0.5
):

    log_val, log_val_t = get_kernels(
        vstate._apply_fun,
        target_state._apply_fun,
        vstate.parameters,
        target_state.parameters,
        samples,
        samples_t,
        vstate.model_state,
        target_state.model_state,
    )
    if weights is None:
        weights = 1.0 / samples.shape[0]
    if weights_t is None:
        weights_t = 1.0 / samples_t.shape[0]

    Hloc = jnp.exp(log_val) * jnp.sum(weights_t * jnp.exp(log_val_t))

    if isinstance(target_state, MCState) and isinstance(vstate, MCState):
        Hloc_cv = jnp.exp(log_val + log_val_t).real + cv_coeff * (
            jnp.exp(2 * (log_val + log_val_t).real) - 1
        )
    else:
        Hloc_cv = Hloc

    return Hloc, Hloc_cv


@partial(jax.jit, static_argnames=("afun", "afun_t"))
def infidelity_sampling_inner(
    afun,
    afun_t,
    params,
    params_t,
    model_state,
    model_state_t,
    sigma,
    sigma_t,
    cv_coeff,
):
    N = sigma.shape[-1]

    σ = sigma.reshape(-1, N)
    σ_t = sigma_t.reshape(-1, N)

    log_val, log_val_t = get_kernels(
        afun, afun_t, params, params_t, σ, σ_t, model_state, model_state_t
    )

    if cv_coeff is not None:
        kernel_vals = jnp.exp(log_val + log_val_t).real + cv_coeff * (
            jnp.exp(2 * (log_val + log_val_t).real) - 1
        )
    else:
        kernel_vals = jnp.exp(log_val) * jnp.mean(jnp.exp(log_val_t))

    mean = jnp.mean(kernel_vals)
    variance = jnp.var(kernel_vals)
    error = jnp.sqrt(variance / kernel_vals.shape[-1])

    I_stats = Stats(
        mean=1 - mean,
        error_of_mean=error,
        variance=variance,
    )

    return I_stats



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from collections.abc import Callable
from collections.abc import Hashable

from netket.jax import canonicalize_dtypes
from netket.utils import struct
from netket.utils.types import DType, PyTree, Array

from netket.hilbert import AbstractHilbert
from netket.operator import AbstractOperator


class ContinuousOperator(AbstractOperator, struct.Pytree):
    r"""This class is the abstract base class for operators defined on a
    continuous Hilbert space. Users interested in implementing new
    quantum Operators for continuous Hilbert spaces should subclass
    `ContinuousOperator` and implement its interface.
    """

    _hilbert: AbstractHilbert = struct.static_field()
    _dtype: DType = struct.static_field()

    def __init__(self, hilbert: AbstractHilbert, dtype: DType | None = None):
        r"""
        Constructs the continuous operator acting on the given hilbert space and
        with a certain data type.

        Args:
            hilbert: The underlying Hilbert space on which the operator is defined
            dtype: Data type of the operator, which is used to infer the dtype of
                expectation values
        """
        dtype = canonicalize_dtypes(float, dtype=dtype)
        self._dtype = dtype
        super().__init__(hilbert)

    @property
    def dtype(self) -> DType:
        return self._dtype

    @abc.abstractmethod
    def _expect_kernel(
        self,
        logpsi: Callable,
        params: PyTree,
        x: Array,
    ) -> Array:
        r"""This method defines the action of the local operator on a given quantum state
        `logpsi` for a given configuration `x`.
        :math:`O_{loc}(x) =  \frac{\bra{x}O{\ket{\psi}}{\bra{x}\ket{\psi}}`
        This method is executed inside of a `jax.jit` block.

        .. note::
            The operator itself must be a Pytree.

        Args:
            logpsi: variational state
            params: parameters for the variational state
            x: a sample of particle positions
        """

    @abc.abstractproperty
    def _attrs(self) -> tuple[Hashable, ...]:
        """This must return a tuple of (hashable) attributes used to compare two operators of
        the same type and that is hashed to compute the hash of the operator.

        To hash arrays you can return them as `nk.utils.HashableArray`.
        """

    @struct.property_cached(pytree_ignore=True)
    def _hash(self) -> int:
        return hash((type(self), self._attrs))

    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, other) -> bool:
        if isinstance(other, type(self)):
            return self._attrs == other._attrs
        return False

    def __add__(self, other):
        if isinstance(self, ContinuousOperator) and isinstance(
            other, ContinuousOperator
        ):
            from netket.operator import SumOperator

            return SumOperator(self, other)
        else:
            return NotImplemented  # pragma: no cover

    def __rmul__(self, other):
        if isinstance(self, ContinuousOperator) and isinstance(other, float):
            return self * other
        else:
            return NotImplemented  # pragma: no cover

    def __mul__(self, other):
        if isinstance(other, float):
            from netket.operator import SumOperator

            return SumOperator(self, coefficients=other)
        else:
            return NotImplemented  # pragma: no cover

    def __sub__(self, other):
        if isinstance(other, ContinuousOperator):
            return self + (-other)
        else:
            return NotImplemented  # pragma: no cover

    def __neg__(self):
        return -1.0 * self



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Internal utility functions to support jax sharding natively within netket.
All functions in here are not part of the public API, internal, and may change without warning.
"""

from functools import partial

import numpy as np

import jax
import jax.numpy as jnp


safe_zip = partial(zip, strict=True)
_identity = lambda x: x


def pad_axis_for_sharding(
    array: jax.Array, *, axis: int = 0, padding_value: float | jax.Array = 0
) -> jax.Array:
    """
    Pads an array along an axis to make it divisible by the number of processes.

    Args:
        array: The array to pad.
        axis: The axis along which to pad.
        padding_value: The value to use for padding.

    Returns:
        The padded array.
    """
    axis_size = array.shape[axis]
    n_devices = jax.device_count()

    if axis_size % n_devices != 0:
        padded_axis_size = int(n_devices * np.ceil(axis_size / n_devices))
        padding_shape = [(0, 0) for _ in range(array.ndim)]
        padding_shape[axis] = (0, padded_axis_size - axis_size)

        array = jnp.pad(
            array,
            padding_shape,
            constant_values=padding_value,
        )
    return array



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Command-line interface for the citation system.
"""

import argparse

from .core import cite


def cite_cli() -> None:
    """
    Command-line interface for the citation system.

    This function is called by the 'netket.cite' console script.
    """
    parser = argparse.ArgumentParser(
        description="Generate citation information for NetKet and associated algorithms",
        prog="netket.cite",
    )
    parser.add_argument(
        "--bib",
        nargs="?",
        const="references.bib",
        help="Generate a BibTeX file. Optionally specify filename (default: references.bib)",
    )

    args = parser.parse_args()

    # Import NetKet to register all citations from decorated functions
    import netket  # noqa: F401

    if args.bib:
        # Automatically add .bib extension if not present (case-insensitive)
        filename = args.bib
        if not filename.lower().endswith(".bib"):
            filename += ".bib"
        cite(output_file=filename, from_cli=True)
    else:
        cite(from_cli=True)



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Expectation value calculations for DeterminantVariationalState using Wick's theorem.
"""
from typing import Union
from functools import partial

import jax
import jax.numpy as jnp

from netket.stats import Stats
from netket.operator import FermionOperator2nd
from netket.vqs import expect, expect_and_grad, expect_and_forces
from netket.vqs.mc.common import force_to_grad

from netket._src.vqs.fermion_mf.state import DeterminantVariationalState

# Import PNC operators from _src to avoid circular imports
from netket._src.operator.particle_number_conserving_fermionic.operators import (
    ParticleNumberConservingFermioperator2nd,
    ParticleNumberAndSpinConservingFermioperator2nd,
)


@partial(jax.jit, static_argnames=("model", "method"))
def jit_apply(model, *args, method):
    return model.apply(*args, method=method)


@partial(jax.jit, static_argnames=("n_body",))
def _compute_wick_determinants(
    rho: jax.Array,
    creation_indices: jax.Array,
    destruction_indices: jax.Array,
    n_body: int,
):
    """
    Compute Wick contractions using determinants for normal-ordered operators.

    For normal-ordered operators c^†_i1 c^†_i2 ... c_j1 c_j2 ..., the expectation
    value is det(rho[j, i]) where rho[j, i] are the density matrix elements
    connecting creation indices i to destruction indices j.

    Args:
        rho: One-body density matrix (n_modes, n_modes)
        creation_indices: Array of shape (M, B) with creation operator indices
        destruction_indices: Array of shape (M, B) with destruction operator indices
        n_body: Number of body operators (B)

    Returns:
        Array of shape (M,) with determinants for each term
    """
    # Build M x B x B tensor where each BxB matrix is rho[creation, destruction]
    # For each term m, we need rho[creation_indices[m, :], destruction_indices[m, :]]
    # This forms a B x B matrix whose determinant gives the Wick contraction
    #
    # For normal-ordered operators c†_i1 c†_i2 ... c_j1 c_j2 ..., Wick's theorem gives:
    # <c†_i1 c†_i2 ... c_j1 c_j2 ...> = det(M)
    # where M[k,l] = <c†_ik c_jl> = rho[jl, ik]
    #
    # So the matrix entry [k, l] should be rho[destruction[l], creation[k]]

    # Use advanced indexing to build the contraction matrices
    # Shape: (M, B, B) where entry [m, k, l] = rho[destruction_indices[m, l], creation_indices[m, k]]
    contraction_matrices = rho[
        destruction_indices[:, None, :], creation_indices[:, :, None]
    ]

    # Compute determinant for each M matrix
    # For 1-body (B=1), determinant is just the single element
    # For 2-body (B=2), we can use the explicit formula
    # For higher bodies, use jnp.linalg.det

    if n_body == 1:
        # Shape (M, 1, 1) -> (M,)
        return contraction_matrices[:, 0, 0]
    elif n_body == 2:
        # Explicit 2x2 determinant for efficiency: ad - bc
        return (
            contraction_matrices[:, 0, 0] * contraction_matrices[:, 1, 1]
            - contraction_matrices[:, 0, 1] * contraction_matrices[:, 1, 0]
        )
    else:
        # General case: use JAX determinant
        return jax.vmap(jnp.linalg.det)(contraction_matrices)


def _expectation_from_rdm(rdm: jax.Array, operator: FermionOperator2nd):
    """
    Compute expectation value using Wick's theorem for normal-ordered operators.

    For a mean-field state with one-body density matrix rdm, this computes
    the expectation value of a fermionic operator by evaluating Wick contractions
    using determinants. This exploits normal ordering for efficient computation.

    Args:
        rdm: The 1-body RDM
        operator: The fermionic operator (must be normal ordered)

    Returns:
        Stats object with the expectation value (variance is 0, since deterministic)
    """
    if operator.order != "N":
        raise NotImplementedError(
            "Only normal-ordered operators are supported."
            " Please convert the operator to normal order using operator.to_normal_order()."
        )

    operators_dict = operator._operators

    # Group terms by number of creation-destruction pairs
    # Dictionary: n_body -> list of (creation_indices, destruction_indices, weight)
    grouped_terms = {}

    for term, weight in operators_dict.items():
        if len(term) == 0:
            # Constant term (identity operator) - handle separately
            grouped_terms.setdefault(0, []).append((None, None, weight))
            continue

        # Since we're in normal order, all creation operators come first
        # Find the split point where dagger changes from 1 to 0
        n_creation = 0
        for orbital, dagger in term:
            if dagger == 1:
                n_creation += 1
            else:
                break

        n_destruction = len(term) - n_creation

        # Only keep terms where number of creation == number of destruction
        if n_creation != n_destruction:
            continue

        n_body = n_creation  # Number of body operators

        # Extract creation and destruction indices
        creation_indices = [term[i][0] for i in range(n_creation)]
        destruction_indices = [term[i][0] for i in range(n_creation, len(term))]

        grouped_terms.setdefault(n_body, []).append(
            (creation_indices, destruction_indices, weight)
        )

    # Compute expectation value for each group
    expectation = 0.0

    # Handle identity term (0-body)
    if 0 in grouped_terms:
        for _, _, weight in grouped_terms[0]:
            expectation += weight

    # Handle n-body terms using JAX
    for n_body, terms in grouped_terms.items():
        if n_body == 0:
            continue

        # Convert to arrays
        creation_array = jnp.array(
            [t[0] for t in terms], dtype=jnp.int32
        )  # Shape: (M, n_body)
        destruction_array = jnp.array(
            [t[1] for t in terms], dtype=jnp.int32
        )  # Shape: (M, n_body)
        weights = jnp.array([t[2] for t in terms])  # Shape: (M,)

        # Compute Wick contractions using determinants
        determinants = _compute_wick_determinants(
            rdm, creation_array, destruction_array, n_body
        )

        # Sign correction for n_body >= 2:
        # For normal-ordered operators with n_body >= 2, the determinant formula
        # det(M) where M[k,l] = <c†_{creation[k]} c_{destruction[l]}> = rdm[destruction[l], creation[k]]
        # gives the opposite sign from the correct Wick contraction result.
        #
        # This has been verified empirically against both the recursive Wick implementation
        # and FullSumState calculations. The sign issue appears regardless of whether indices
        # are sorted or unsorted, suggesting it's a fundamental aspect of how the determinant
        # formula relates to normal-ordered expectation values for n >= 2.
        #
        # For 1-body operators (n=1), the determinant (single matrix element) gives the
        # correct sign without modification.
        if n_body >= 2:
            determinants = -determinants

        # Sum weighted contributions
        expectation += jnp.sum(weights * determinants)

    # Return as Stats object with zero variance (deterministic)
    # Note: expectation might be a JAX tracer, so we can't convert to Python complex
    return Stats(mean=expectation, variance=0.0, error_of_mean=0.0)


@expect.register
def expect_fermionicmf(
    vstate: DeterminantVariationalState,
    operator: FermionOperator2nd,
):
    return _expectation_from_rdm(vstate.rdm, operator)


@expect_and_grad.register
def expect_and_grad_fermionicmf(
    vstate: DeterminantVariationalState,
    operator: FermionOperator2nd,
    *args,
    use_covariance: bool | None = None,
    mutable,
    **kwargs,
):
    Ō, Ō_forces = expect_and_forces(vstate, operator, *args, mutable=mutable, **kwargs)
    Ō_grad = force_to_grad(Ō_forces, vstate.parameters)
    return Ō, Ō_grad


@expect.register
def expect_fermionicmf_pnc(
    vstate: DeterminantVariationalState,
    operator: Union[
        ParticleNumberConservingFermioperator2nd,
        ParticleNumberAndSpinConservingFermioperator2nd,
    ],
):
    fermion_op = operator.to_fermionoperator2nd().to_normal_order()
    return expect_fermionicmf(vstate, fermion_op)


@expect_and_grad.register
def expect_and_grad_fermionicmf_pnc(
    vstate: DeterminantVariationalState,
    operator: Union[
        ParticleNumberConservingFermioperator2nd,
        ParticleNumberAndSpinConservingFermioperator2nd,
    ],
    *args,
    use_covariance: bool | None = None,
    mutable,
    **kwargs,
):
    Ō, Ō_forces = expect_and_forces(vstate, operator, *args, mutable=mutable, **kwargs)
    Ō_grad = force_to_grad(Ō_forces, vstate.parameters)
    return Ō, Ō_grad


def _expectation_and_forces_from_rdm(
    vstate: DeterminantVariationalState,
    operator: FermionOperator2nd,
    mutable,
):
    """
    Compute expectation value and forces using Wick's theorem.

    For a deterministic state, the forces are defined as half the gradient,
    so that force_to_grad correctly converts them to the full gradient.

    Returns:
        Stats object with expectation value and forces (half-gradients)
    """
    model = vstate._model
    model_state = vstate.model_state

    def expectation_fun(params):
        rdm = jit_apply(
            model, {"params": params, **model_state}, method=model.rdm_one_body
        )
        exp_val = _expectation_from_rdm(rdm, operator).mean
        # For hermitian operators, expectation should be real
        return jnp.real(exp_val), rdm

    (exp_val, rdm), grad = jax.value_and_grad(
        expectation_fun, holomorphic=False, has_aux=True
    )(vstate.parameters)
    if vstate._rdm is None:
        vstate._rdm = rdm

    forces = jax.tree_util.tree_map(lambda g: 0.5 * jnp.conj(g), grad)
    stats = Stats(mean=exp_val, variance=0.0, error_of_mean=0.0)
    return stats, forces


@expect_and_forces.register
def expect_and_forces_fermionicmf(
    vstate: DeterminantVariationalState,
    operator: FermionOperator2nd,
    *args,
    mutable,
    **kwargs,
):
    return _expectation_and_forces_from_rdm(vstate, operator, mutable)


@expect_and_forces.register
def expect_and_forces_fermionicmf_pnc(
    vstate: DeterminantVariationalState,
    operator: Union[
        ParticleNumberConservingFermioperator2nd,
        ParticleNumberAndSpinConservingFermioperator2nd,
    ],
    *args,
    mutable,
    **kwargs,
):
    fermion_op = operator.to_fermionoperator2nd().to_normal_order()
    return expect_and_forces_fermionicmf(
        vstate, fermion_op, *args, mutable=mutable, **kwargs
    )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netket.utils.mpi.mpi import (
    mpi4jax_available as _depr_available,
    MPI as _depr_MPI,
    MPI_py_comm as _depr_MPI_py_comm,
    MPI_jax_comm as _depr_MPI_jax_comm,
    n_nodes as _depr_n_nodes,
    node_number as _depr_node_number,
    rank as _depr_rank,
)

from netket.utils.mpi.primitives import (
    mpi_all as _depr_mpi_all,
    mpi_allgather as _depr_mpi_allgather,
    mpi_any as _depr_mpi_any,
    mpi_bcast as _depr_mpi_bcast,
    mpi_max as _depr_mpi_max,
    mpi_mean as _depr_mpi_mean,
    mpi_sum as _depr_mpi_sum,
    mpi_all_jax as _depr_mpi_all_jax,
    mpi_allgather_jax as _depr_mpi_allgather_jax,
    mpi_any_jax as _depr_mpi_any_jax,
    mpi_bcast_jax as _depr_mpi_bcast_jax,
    mpi_max_jax as _depr_mpi_max_jax,
    mpi_mean_jax as _depr_mpi_mean_jax,
    mpi_sum_jax as _depr_mpi_sum_jax,
    mpi_gather as _depr_mpi_gather,
    mpi_gather_jax as _depr_mpi_gather_jax,
    mpi_alltoall_jax as _depr_mpi_alltoall_jax,
    mpi_reduce_sum_jax as _depr_mpi_reduce_sum_jax,
    mpi_allreduce_sum_jax as _depr_mpi_allreduce_sum_jax,
    mpi_scatter_jax as _depr_mpi_scatter_jax,
)

from netket.utils.deprecation import deprecation_getattr as _deprecation_getattr

msg = (
    "netket.utils.mpi is deprecated, as MPI support has been removed. "
    "Refer to `jax.process_count()` and `jax.process_index()` for process count and rank. "
    "And to other jax functions for distributed operations."
)
_deprecations = {
    # July 2025, NetKet 3.19
    "available": (msg, _depr_available),
    "MPI": (msg, _depr_MPI),
    "MPI_py_comm": (msg, _depr_MPI_py_comm),
    "MPI_jax_comm": (msg, _depr_MPI_jax_comm),
    "n_nodes": (msg, _depr_n_nodes),
    "node_number": (msg, _depr_node_number),
    "rank": (msg, _depr_rank),
    "mpi_all": (msg, _depr_mpi_all),
    "mpi_allgather": (msg, _depr_mpi_allgather),
    "mpi_any": (msg, _depr_mpi_any),
    "mpi_bcast": (msg, _depr_mpi_bcast),
    "mpi_max": (msg, _depr_mpi_max),
    "mpi_mean": (msg, _depr_mpi_mean),
    "mpi_sum": (msg, _depr_mpi_sum),
    "mpi_all_jax": (msg, _depr_mpi_all_jax),
    "mpi_allgather_jax": (msg, _depr_mpi_allgather_jax),
    "mpi_any_jax": (msg, _depr_mpi_any_jax),
    "mpi_bcast_jax": (msg, _depr_mpi_bcast_jax),
    "mpi_max_jax": (msg, _depr_mpi_max_jax),
    "mpi_mean_jax": (msg, _depr_mpi_mean_jax),
    "mpi_sum_jax": (msg, _depr_mpi_sum_jax),
    "mpi_gather": (msg, _depr_mpi_gather),
    "mpi_gather_jax": (msg, _depr_mpi_gather_jax),
    "mpi_alltoall_jax": (msg, _depr_mpi_alltoall_jax),
    "mpi_reduce_sum_jax": (msg, _depr_mpi_reduce_sum_jax),
    "mpi_allreduce_sum_jax": (msg, _depr_mpi_allreduce_sum_jax),
    "mpi_scatter_jax": (msg, _depr_mpi_scatter_jax),
}

__getattr__ = _deprecation_getattr(__name__, _deprecations)

del _deprecation_getattr



# Copyright 2020, 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Display random tips when NetKet is imported interactively."""

import random
import sys

TIPS = [
    # Drivers
    "Prefer the new nk.driver.VMC_SR over VMC which supports minSR and SPRING.",
    "Use driver.run(..., timeit=True) to know where your dominant cost is.",
    # Sampling
    "With many Markov Chains (e.g GPUs), n_discard_per_chain>5 is often inefficient.",
    "On GPUs an efficient choice is to run ~512 Markov Chains per GPU.",
    r"If timeit=True signals high \% spent sampling n_discarded, consider lowering it.",
    # Operators
    "Avoid LocalOperator with have N-body operators (N>4). PauliStrings is ok.",
    # Loggers
    "You can load logged data with nk.utils.history.HistoryDict.from_file(data.log).",
    "You can plot data with JsonLog.data['Energy'][:-30].plot().",
    # Variational states
    "To build H|ψ⟩ use nk.vqs.apply_operator(H, vstate_ψ).",
    "log(ψ) ∈ ℜ → ψ = exp(logψ) ∈ ℜ₊ (real NN gives only positive wave-function).",
    "You can use flax.linen, flax.nnx and equinox to define neural networks.",
    # utils
    "You must cite NetKet according to our policy. Use nk.cite() to find out how.",
    "Debug multi-node HPC? `djaxrun -np 2 python Examples/Sharding/multi_process.py`",
    # extras
    "uv is a replacement for pip which helps you follow good software practices.",
    "You can disable these tips by setting export NETKET_NO_TIPS=1 in your .bashrc.",
    "😍 Love? 😡 Hate? these tips? Let us know at https://forms.gle/ej8eDGFRu2mww5mQ9",
]


def _is_interactive():
    """Check if Python is running in an interactive environment."""
    try:
        # Check if we're in IPython/Jupyter
        ipython = get_ipython()  # noqa: F821
        # IPython is available, but check if we're running a script
        # In IPython/Jupyter interactive mode, __file__ is not defined in the main namespace
        # We check the config to see if it's truly interactive
        if hasattr(ipython, "config"):
            # If IPython is running a file/script, we're not interactive
            if hasattr(ipython, "user_ns") and "__file__" in ipython.user_ns:
                return False
        # Check if we're in a Jupyter notebook or IPython shell
        return ipython.__class__.__name__ in [
            "TerminalInteractiveShell",
            "ZMQInteractiveShell",
        ]
    except NameError:
        # Check if we're in a regular Python REPL
        return hasattr(sys, "ps1")


def show_random_tip():
    """Display a random tip using rich formatting if in interactive mode."""
    if not _is_interactive():
        return

    from netket.utils import config

    if config.netket_no_tips:
        return

    import textwrap
    from rich.console import Console

    tip = random.choice(TIPS)
    console = Console()

    # Wrap text with hanging indent
    prefix = "∣NK⟩ Tip: "
    width = console.width
    wrapped = textwrap.fill(
        tip,
        width=width,
        initial_indent=prefix,
        subsequent_indent=" " * len(prefix),
    )

    # Apply styling only to the prefix
    lines = wrapped.split("\n")
    for i, line in enumerate(lines):
        if i == 0:
            # First line: style the prefix
            console.print(f"[bold cyan]{prefix}[/bold cyan]{line[len(prefix):]}")
        else:
            # Subsequent lines: already have proper indentation
            console.print(line)



"""
This is a script to run a distributed job using JAX on a single node.
Useful for testing with pytest under jax distributed.

It works by pretending it's running under SLURM, setting up the environment variables
and launching the command in multiple processes.
"""

import argparse
import os
import random
import socket
import sys
import threading
import re
import subprocess
from rich.live import Live
from rich.panel import Panel
from rich.text import Text
from rich.console import Group


def find_free_port():
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("localhost", 0))
        return s.getsockname()[1]


def render_panels(buffers):
    panels = []
    for rank, content in enumerate(buffers):
        text = Text.from_ansi(content.strip()[-1000:])
        panels.append(Panel(text, title=f"[p{rank}]", border_style="green"))
    return Group(*panels)


def main():
    parser = argparse.ArgumentParser(
        description="Distributed runner with live panels and logs, to run a jax.ditributed job on a single node. ",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("-np", type=int, help="Number of processes to run")
    parser.add_argument(
        "-o", type=str, default="djaxrunlog", help="Base output filename prefix"
    )
    parser.add_argument(
        "-s",
        "--simple",
        action="store_true",
        help="Use simple output mode (no Rich panels)",
    )
    parser.add_argument(
        "-f",
        "--filter",
        type=int,
        default=-1,
        help="Filter output to only show output from the process of this rank (negative values means all ranks)",
    )
    parser.add_argument("command", nargs=argparse.REMAINDER, help="Command to run")

    args = parser.parse_args()

    # Show help if no arguments or missing required ones
    if args.np is None or not args.command:
        parser.print_help()
        sys.exit(1)

    num_procs = args.np
    command = args.command
    outprefix = args.o
    simple_output = args.simple
    filter_output = args.filter

    port = find_free_port()
    job_id = random.randint(100000, 999999)
    node_list = ",".join(["127.0.0.1"] * num_procs)

    buffers = ["" for _ in range(num_procs)]
    threads = []
    procs = []
    log_files = []
    ansi_escape = re.compile(r"\x1b\[[0-9;?]*[a-zA-Z]")

    def start_proc(rank):
        env = os.environ.copy()
        env.update(
            {
                "JAX_COORDINATOR_ADDRESS": f"127.0.0.1:{port}",
                "SLURM_JOB_ID": str(job_id),
                "SLURM_NTASKS": str(num_procs),
                "SLURM_PROCID": str(rank),
                "SLURM_LOCALID": str(rank),
                "SLURM_STEP_NODELIST": node_list,
                "PYTHONUNBUFFERED": "1",
                "DJAXRUN": "1",
            }
        )

        logfile = open(f"{outprefix}-p{rank}", "w", buffering=1)
        log_files.append(logfile)

        proc = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env=env,
            bufsize=1,
            text=True,
        )
        procs.append(proc)

        def reader():
            for line in proc.stdout:
                buffers[rank] += line
                cleaned = ansi_escape.sub("", line)
                logfile.write(cleaned)
                if simple_output and (filter_output < 0 or rank == filter_output):
                    print(f"[p{rank}] {line.strip()}")

            proc.stdout.close()
            logfile.close()

        t = threading.Thread(target=reader)
        t.daemon = True
        t.start()
        threads.append(t)

    for rank in range(num_procs):
        start_proc(rank)

    try:
        if simple_output:
            for t in threads:
                t.join()
        else:
            with Live(
                render_panels(buffers), refresh_per_second=10, screen=False
            ) as live:
                while any(t.is_alive() for t in threads):
                    live.update(render_panels(buffers))
                    for t in threads:
                        t.join(timeout=0.1)
    except KeyboardInterrupt:
        print("\n[!] Ctrl-C detected, terminating processes...", file=sys.stderr)
    finally:
        # First, try to terminate all still-running processes
        for proc in procs:
            if proc.poll() is None:
                try:
                    proc.terminate()
                except Exception:
                    pass
        # Now wait for all to exit
        for proc in procs:
            try:
                proc.wait(timeout=5)
            except Exception:
                pass
        # Close all log files
        for logfile in log_files:
            if not logfile.closed:
                logfile.close()

        # Check exit codes
        exit_codes = [proc.returncode for proc in procs]
        print(f"Exit codes of the processes where {exit_codes}")
        if any(code != 0 for code in exit_codes):
            sys.exit(1)


if __name__ == "__main__":
    main()



from netket._src.symmetry.representation import Representation
from netket._src.symmetry.canonical_representation import (
    canonical_representation,
)

from . import group

from netket.utils import _auto_export

_auto_export(__name__)



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import cached_property

import numpy as np

from netket.utils.group import Element, FiniteGroup
from netket.operator import DiscreteJaxOperator, SumOperator
from netket.vqs.mc.mc_state.state import MCState


class Representation:
    """
    A representation of a group.

    This is the central object for symmetrizing wavefunctions. The character
    table of the group of the representation indicates the various subspaces on
    which the wavefunction can be projected. A variational state :code:`vstate`
    can be projected on the subspace associated to the character :code:`k` by
    calling :code:`representation.project(vstate, k)` , which returns the new,
    projected variational state.

    For more details on the theory of representations of finite group and its
    application in variational Monte Carlo, we refer to :doc:`/advanced/symmetry`.

    For a tutorial of symmetrizing variational states in practice, we refer to
    the :doc:`/tutorials/symmetry_tutorial`.
    """

    def __init__(
        self,
        group: FiniteGroup,
        representation_dict: dict[Element, DiscreteJaxOperator],
    ):
        """
        Construct a Representation.

        Args:
            group: The group being represented.
            representation_dict: The dictionary that defines the representation.
                Its keys must be the elements of :code:`group`. The value this
                dictionary associates to a group element is the operator that
                the representation maps this group element onto.
        """

        if not isinstance(group, FiniteGroup):
            raise TypeError("group must be a FiniteGroup")

        operator = next(iter(representation_dict.values()))
        hilbert = operator.hilbert

        for element, operator in representation_dict.items():
            assert hilbert == operator.hilbert
            assert element in group.elems

        operators = tuple(representation_dict[el] for el in group.elems)

        if not len(group.elems) == len(representation_dict):
            raise ValueError(
                "The representation dictionary must have the same "
                "number of elements as the group."
            )

        self.hilbert = hilbert
        self.group = group
        self.operators = operators

    def __repr__(self):
        return f"Representation(group={self.group}, hilbert={self.hilbert})"

    def __getitem__(self, key):
        if isinstance(key, Element):
            return self.representation_dict[key]
        elif isinstance(key, int):
            return self.representation_dict[self.group[key]]
        raise TypeError("Index should be integer or group element")

    def __hash__(self):
        return hash(("Representation", self.hilbert, self.group, self.operators))

    def __eq__(self, other):
        if type(self) is type(other):
            return (
                self.hilbert == other.hilbert
                and self.group == other.group
                and self.operators == other.operators
            )
        return False

    @cached_property
    def representation_dict(self) -> dict[Element, DiscreteJaxOperator]:
        """
        Dictionary associating every group element to a representation

        Equivalent to `{el: rep for (el, rep) in (self.group.elems,
        self.operators)}`
        """
        return {el: rep for (el, rep) in zip(self.group.elems, self.operators)}

    def __iter__(self):
        return zip(self.group.elems, self.operators, strict=True)

    def __len__(self):
        return len(self.representation_dict)

    def projector(
        self, character_index: int, *, atol: float = 1e-15
    ) -> DiscreteJaxOperator:
        """Build the projection operator corresponding to a given
        irreducible representation.

        As there might be some terms in the projector whose character is 0,
        we drop them by default according to a given absolute tolerance.

        Args:
            character_index: The index of the irreducible representation for which to
                construct the projector.
            atol: Absolute tolerance used to drop zero-terms from the projector to reduce
                its number of connected entries (defaults to 1e-15).

        Return:
            A `netket.operator.DiscreteJaxOperator` that projects on the relevant subspace.
        """
        character_table = self.group.character_table()
        prefactor = character_table[character_index, 0] / len(self.group.elems)

        # Build manually the SumOperator for efficiency when operating with
        # large groups
        operators = np.array([self[g] for g in self.group], dtype=object)

        # Filter out the characters that vanish (do before normalizing to avoid even smaller values)
        mask = ~np.isclose(np.conj(character_table[character_index]), 0.0, atol=atol)
        operators = np.array([self[g] for g in self.group], dtype=object)
        coefficients = prefactor * np.conj(character_table[character_index])

        operators = operators[mask]
        coefficients = coefficients[mask]

        projector = SumOperator(*operators, coefficients=coefficients)
        return projector

    def project(self, state, character_index: int, *, atol: float = 1e-15) -> MCState:
        """Return the state projected onto the subspace associated to the
        irreducible representation specified by character_index.
        """
        from netket._src.vqs.transformed_vstate import apply_operator

        projector = self.projector(character_index, atol=atol)
        projected_state = apply_operator(projector, state)
        return projected_state

    @property
    def character(self) -> np.ndarray:
        """
        The vector storing the character of the representation.

        Corresponds to ``[op.trace() for (_, op) in self]``.

        Requires that each operator of the representation
        implements the trace method.
        """
        try:
            character = [op.trace() for perm, op in self]
        except NotImplementedError as err:
            raise NotImplementedError(
                "At least one operator of the representation "
                "does not implement the trace. The original error "
                "follows:"
            ) from err
        return np.array(character)

    def irrep_subspace_dims(self) -> int:
        """
        Return the dimension of the subspace associated to each irreducible
        representation.

        Requires that each operator of the representation implements
        the trace method.
        """
        character_table = self.group.character_table()
        group_order = len(self.group.elems)

        irrep_count = character_table @ self.character / group_order
        irrep_dims = np.round(irrep_count * character_table[:, 0]).astype(int)
        return irrep_dims

    def symmetry_adapted_basis(self) -> tuple[np.ndarray, np.ndarray]:
        """
        Return a tuple `(mat, irrep_dims)`, where `mat` is the change of basis
        matrix associated to a symmetry adapted basis, and `irrep_dims` is
        an array giving the dimension of the subspace associated to each
        irreducible representation.

        The basis vectors of `mat` are ordered by the index of their irreducible
        representation.
        """
        n_irreps = self.group.character_table().shape[0]
        projectors = [self.projector(k).to_dense() for k in range(n_irreps)]

        cob_matrix = np.zeros(
            (self.hilbert.n_states, self.hilbert.n_states), dtype=complex
        )
        current_index = 0

        irrep_dims = []

        for projector in projectors:

            eigvals, eigvecs = np.linalg.eig(projector)
            is_in_image = np.abs(eigvals - 1) < 1e-12
            selected_eigvecs = eigvecs[:, is_in_image]

            projector_rank = selected_eigvecs.shape[1]
            irrep_dims.append(projector_rank)

            cob_matrix[:, current_index : current_index + projector_rank] = (
                selected_eigvecs
            )
            current_index += projector_rank

        return cob_matrix, np.array(irrep_dims)


# This is work in progress

# def is_commuting(self, other):
#     """
#     Check whether all operators of one representation commute with all operators of
#     the other.
#     """
#     for operator_1 in self.representation_dict.values():
#         for operator_2 in other.representation_dict.values():
#             assert operator_1 @ operator_2 - operator_2 @ operator_1 == 0

# # We might want a fast mode where only the fast checks are made,
# # and an exhaustive mode where all checks are made.
# def check_representation(self):
#     """
#     Check whether the representation is valid by checking that the
#     representation properties are satisfied.
#     """

#     is_representation = True

#     # Identity property
#     for g in self.group:
#         if isinstance(g, Identity):
#             if (
#                 not jnp.linalg.norm(
#                     self[g].to_dense - jnp.eye(self.hilbert_space.n_states)
#                 )
#                 < 1e-14
#             ):
#                 is_representation = False

#     # Compatibility property
#     for g_1, g_2 in product(self.group, self.group):
#         product_inside = self[g_1 @ g_2]
#         product_outside = self[g_1] @ self[g_2]
#         if not product_inside == product_outside:
#             is_representation = False

#     return is_representation



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from jax.tree_util import register_pytree_node_class

from netket.hilbert import AbstractHilbert
from netket.operator import DiscreteJaxOperator
from netket.jax import canonicalize_dtypes
from netket.utils.group import Permutation, Identity
from netket.utils.types import DType


@register_pytree_node_class
class PermutationOperatorBase(DiscreteJaxOperator):
    """
    Base class for permutation operators acting on an
    Hilbert space. Used to express lattice symmetries and symmetrize
    variational states with respect to these symmetries.

    Subclasses implement this on Spin, Fermionic or other hilbert spaces.
    Look at :class:`netket.operator.permutation.PermutationOperator` and
    :class:`netket.operator.permutation.PermutationOperatorFermion` for
    examples.

    For mathematical details on the definition of a permutation operator
    and its justification, we refer to :doc:`/advanced/symmetry`.
    """

    def __init__(
        self,
        hilbert: AbstractHilbert,
        permutation: Permutation,
        *,
        dtype: DType | None = float,
    ):
        """
        Construct the permutation operator.

        Args:
            hilbert: The Hilbert space.
            permutation: The permutation that defines the operator as
                specified in :doc:`/advanced/symmetry`.
        """
        if isinstance(permutation, Identity):
            permutation = Permutation(
                permutation_array=np.arange(hilbert.size),
                name="Identity",
                validate=False,
            )

        if not isinstance(permutation, Permutation):
            raise TypeError("permutation must be a Permutation object.")
        if not hilbert.size == permutation.permutation_array.size:
            raise ValueError(
                "Permutation size does not correspond to Hilbert space size."
            )

        dtype = canonicalize_dtypes(dtype=dtype)

        super().__init__(hilbert)
        self.permutation = permutation
        self._dtype = dtype

    def __repr__(self):
        if self.permutation._name is not None:
            perm_repr = f"{self.permutation._name}:{self.permutation.permutation_array}"
        else:
            perm_repr = f"{self.permutation.permutation_array}"
        return f"{type(self).__name__}({self.hilbert}, {perm_repr}, dtype={self.dtype})"

    def tree_flatten(self):
        struct_data = {
            "hilbert": self.hilbert,
            "permutation": self.permutation,
            "dtype": self.dtype,
        }
        return (), struct_data

    @classmethod
    def tree_unflatten(cls, struct_data, array_data):
        return cls(**struct_data)

    @property
    def max_conn_size(self) -> int:
        return 1

    @property
    def dtype(self):
        return self._dtype

    def __hash__(self):
        return hash((self.hilbert, self.permutation))

    def __eq__(self, other):
        if type(self) is type(other):
            return (
                self.hilbert == other.hilbert and self.permutation == other.permutation
            )
        return False

    def __matmul__(self, other):
        if type(self) is type(other) and self.hilbert == other.hilbert:
            return type(self)(self.hilbert, self.permutation @ other.permutation)
        else:
            return super().__matmul__(other)

    def trace(self) -> complex:
        """
        Compute the trace of the operator on the given Hilbert space.

        This could also raise a `NotImplementedError`.
        """
        raise NotImplementedError(
            f"Trace method not implemented for class `{type(self)}`"
        )
